from pytestqt import qtbot
from xicam.gui.widgets.ROI import ArcROI


def test_arcroi(qtbot):
    import pyqtgraph as pg
    import numpy as np

    imageview = pg.ImageView()
    data = np.random.random((10, 10))
    imageview.setImage(data)

    roi = ArcROI(center=(5, 5), radius=5)
    imageview.view.addItem(roi)
    imageview.show()

    # qtbot.addWidget(imageview)

    assert np.sum(roi.getArrayRegion(np.ones((10, 10)))) == 16

    qtbot.waitForWindowShown(imageview)



# -*- coding: utf-8 -*-
from functools import WRAPPER_ASSIGNMENTS
from pyqtgraph import ImageView, InfiniteLine, mkPen, ScatterPlotItem, ImageItem, PlotItem
from qtpy.QtGui import QTransform, QPolygonF
from qtpy.QtWidgets import QLabel, QErrorMessage, QSizePolicy, QPushButton
from qtpy.QtCore import Qt, Signal, Slot, QSize, QPointF, QRectF
import numpy as np

# from pyFAI.geometry import Geometry
from xicam.core import msg
from xicam.core.data import MetaXArray
from xicam.gui.widgets.elidedlabel import ElidedLabel
from xicam.gui.widgets.ROI import BetterPolyLineROI
import enum

from xicam.plugins import manager as pluginmanager
import inspect


# NOTE: PyQt widget mixins have pitfalls; note #2 here: http://trevorius.com/scrapbook/python/pyqt-multiple-inheritance/

# NOTE: PyFAI geometry position vector is: x = up
#                                          y = right
#                                          z = beam


def q_from_angles(phi, alpha, wavelength):
    r = 2 * np.pi / wavelength
    qx = r * np.sin(phi) * np.cos(alpha)
    qy = r * np.cos(phi) * np.sin(alpha)
    qz = r * (np.cos(phi) * np.cos(alpha) - 1)

    return np.array([qx, qy, qz])


def alpha(x, y, z):
    return np.arctan2(y, z)


def phi(x, y, z):
    return np.arctan2(x, z)


class DisplayMode(enum.Enum):
    raw = enum.auto()
    cake = enum.auto()
    remesh = enum.auto()


class PixelSpace(ImageView):
    def __init__(self, *args, **kwargs):
        # Add axes
        self.axesItem = PlotItem()
        self.axesItem.axes["left"]["item"].setZValue(10)
        self.axesItem.axes["top"]["item"].setZValue(10)
        if "view" not in kwargs:
            kwargs["view"] = self.axesItem

        self._transform = QTransform()
        self._raw_image = None

        super(PixelSpace, self).__init__(*args, **kwargs)

        self.imageItem.sigImageChanged.connect(self.updateAxes)

    def transform(self, img=None):
        # Build Quads
        shape = img.shape
        a = [(0, shape[-2] - 1), (shape[-1] - 1, shape[-2] - 1), (shape[-1] - 1, 0), (0, 0)]

        b = [(0, 1), (shape[-1] - 1, 1), (shape[-1] - 1, shape[-2]), (0, shape[-2])]

        quad1 = QPolygonF()
        quad2 = QPolygonF()
        for p, q in zip(a, b):
            quad1.append(QPointF(*p))
            quad2.append(QPointF(*q))

        transform = QTransform()
        QTransform.quadToQuad(quad1, quad2, transform)

        for item in self.view.items:
            if isinstance(item, ImageItem):
                item.setTransform(transform)
        self._transform = transform
        return img, transform

    def setImage(self, img, *args, **kwargs):
        if img is None:
            return

        if getattr(self, "displaymode", DisplayMode.raw) == DisplayMode.raw:
            self._raw_image = img

        if not kwargs.get("transform", None):
            img, transform = self.transform(img)
            self.updateAxes()
            super(PixelSpace, self).setImage(img, *args, transform=transform, **kwargs)

        else:
            super(PixelSpace, self).setImage(img, *args, **kwargs)

    def setTransform(self):
        self.setImage(self._raw_image)  # this should loop back around to the respective transforms

    def updateAxes(self):
        self.axesItem.setLabel("bottom", "x (px)")  # , units='s')
        self.axesItem.setLabel("left", "z (px)")


class QSpace(PixelSpace):
    def __init__(self, *args, geometry=None, **kwargs):
        self.displaymode = DisplayMode.raw
        self._geometry = None  # type: AzimuthalIntegrator

        super(QSpace, self).__init__(*args, **kwargs)

        self.setGeometry(geometry)

    def setGeometry(self, geometry):
        if callable(geometry):
            geometry = geometry()
        self._geometry = geometry
        self.setTransform()


class EwaldCorrected(QSpace):
    def setDisplayMode(self, mode):
        self.displaymode = mode
        if hasattr(self, "drawCenter"):
            self.drawCenter()
        self.setTransform()

    def transform(self, img=None):
        if not self._geometry or not self.displaymode == DisplayMode.remesh:
            return super(EwaldCorrected, self).transform(img)  # Do pixel space transform when not calibrated

        from camsaxs import remesh_bbox

        img, q_x, q_z = remesh_bbox.remesh(np.squeeze(img), self._geometry, reflection=False, alphai=None)

        # Build Quads
        shape = img.shape
        a = shape[-2] - 1, 0  # bottom-left
        b = shape[-2] - 1, shape[-1] - 1  # bottom-right
        c = 0, shape[-1] - 1  # top-right
        d = 0, 0  # top-left

        quad1 = QPolygonF()
        quad2 = QPolygonF()
        for p, q in zip([a, b, c, d], [a, b, c, d]):  # the zip does the flip :P
            quad1.append(QPointF(*p[::-1]))
            quad2.append(QPointF(q_x[q], q_z[q]))

        transform = QTransform()
        QTransform.quadToQuad(quad1, quad2, transform)

        for item in self.view.items:
            if isinstance(item, ImageItem):
                item.setTransform(transform)
        self._transform = transform

        return img, self._transform

    def setImage(self, img, *args, **kwargs):
        if img is None:
            return

        self._raw_image = img

        if self._geometry:
            transform_img, transform = self.transform(img)
            super(EwaldCorrected, self).setImage(transform_img, *args, transform=transform, **kwargs)

        else:
            super(EwaldCorrected, self).setImage(img, *args, **kwargs)

    def updateAxes(self):
        if self.displaymode == DisplayMode.remesh:
            self.axesItem.setLabel("bottom", "q<sub>x</sub> (Å⁻¹)")  # , units='s')
            self.axesItem.setLabel("left", "q<sub>z</sub> (Å⁻¹)")
        else:
            super(EwaldCorrected, self).updateAxes()


class CenterMarker(QSpace):
    def __init__(self, *args, **kwargs):
        self.centerplot = ScatterPlotItem(brush="r")
        self.centerplot.setZValue(100)

        super(CenterMarker, self).__init__(*args, **kwargs)

        self.addItem(self.centerplot)
        self.drawCenter()

    def drawCenter(self):
        try:
            fit2d = self._geometry.getFit2D()
        except (TypeError, AttributeError):
            pass
        else:
            if self.imageItem.image is not None:
                if self.displaymode == DisplayMode.raw:
                    x = fit2d["centerX"]
                    y = self._raw_image.shape[-2] - fit2d["centerY"]
                    self.centerplot.setData(x=[x], y=[y])
                elif self.displaymode == DisplayMode.remesh:
                    self.centerplot.setData(x=[0], y=[0])

    def setGeometry(self, geometry):
        super(CenterMarker, self).setGeometry(geometry)
        self.drawCenter()


class Crosshair(ImageView):
    def __init__(self, *args, **kwargs):
        super(Crosshair, self).__init__(*args, **kwargs)
        linepen = mkPen("#FFA500")
        self._vline = InfiniteLine((0, 0), angle=90, movable=False, pen=linepen)
        self._hline = InfiniteLine((0, 0), angle=0, movable=False, pen=linepen)

        self._vline.setVisible(False)
        self._hline.setVisible(False)

        self.addItem(self._vline)
        self.addItem(self._hline)

        self.scene.sigMouseMoved.connect(self.moveCrosshair)

    def moveCrosshair(self, pos):
        if self.view.getViewBox().sceneBoundingRect().contains(pos):
            mousePoint = self.view.getViewBox().mapSceneToView(pos)
            x, y = mousePoint.x(), mousePoint.y()

            if self.imageItem.mapRectToView(self.imageItem.boundingRect()).contains(mousePoint):  # within bounds
                self._vline.setPos(x)
                self._hline.setPos(y)
                self._hline.setVisible(True)
                self._vline.setVisible(True)
            else:
                self._hline.setVisible(False)
                self._vline.setVisible(False)


class PixelCoordinates(PixelSpace):
    def __init__(self, *args, **kwargs):
        super(PixelCoordinates, self).__init__(*args, **kwargs)

        self._coordslabel = QLabel(
            "<div style='font-size:12pt;background-color:#111111; " "text-overflow: ellipsis; width:100%;'>&nbsp;</div>"
        )

        # def sizeHint():
        #     sizehint = QSize(self.ui.graphicsView.width()-10, self._coordslabel.height())
        #     return sizehint
        # self._coordslabel.sizeHint = sizeHint
        self._coordslabel.setSizePolicy(
            QSizePolicy.Ignored, QSizePolicy.Ignored
        )  # TODO: set sizehint to take from parent, not text
        self.ui.gridLayout.addWidget(self._coordslabel, 2, 0, 1, 1, alignment=Qt.AlignHCenter)

        self.scene.sigMouseMoved.connect(self.displayCoordinates)

    def displayCoordinates(self, pos):
        if self.view.sceneBoundingRect().contains(pos):
            mousePoint = self.view.getViewBox().mapSceneToView(pos)
            pos = QPointF(mousePoint.x(), mousePoint.y())

            if self.imageItem.mapRectToView(self.imageItem.boundingRect()).contains(mousePoint):  # within bounds
                # angstrom=QChar(0x00B5)
                pxpos = self.imageItem.mapFromView(pos)

                self.formatCoordinates(pxpos, pos)
            else:
                self._coordslabel.setText("<div style='font-size:12pt;background-color:#111111;'>&nbsp;</div>")

    def formatCoordinates(self, pxpos, pos):
        """
        when the mouse is moved in the viewer, recalculate coordinates
        """

        try:
            I = self.imageItem.image[int(pxpos.y()), int(pxpos.x())]
        except IndexError:
            I = 0

        self._coordslabel.setText(
            f"<div style='font-size: 12pt;background-color:#111111; color:#FFFFFF;"
            f"text-overflow: ellipsis; width:100%;'>"
            f"x={pxpos.x():0.1f}, "
            f"<span style=''>y={pxpos.y():0.1f}</span>, "
            f"<span style=''>I={I:0.0f}</span></div>"
        )


class QCoordinates(QSpace, PixelCoordinates):
    def formatCoordinates(self, pxpos, pos):
        """
        when the mouse is moved in the viewer, recalculate coordinates
        """

        try:
            I = self.imageItem.image[int(pxpos.y()), int(pxpos.x())]
        except IndexError:
            I = 0
        self._coordslabel.setText(
            f"<div style='font-size: 12pt;background-color:#111111; color:#FFFFFF; "
            f"text-overflow: ellipsis; width:100%;'>"
            f"x={pxpos.x():0.1f}, "
            f"<span style=''>y={self.imageItem.image.shape[-2] - pxpos.y():0.1f}</span>, "
            f"<span style=''>I={I:0.0f}</span>, "
            f"q={np.sqrt(pos.x() ** 2 + pos.y() ** 2):0.3f} \u212B\u207B\u00B9, "
            f"q<sub>z</sub>={pos.y():0.3f} \u212B\u207B\u00B9, "
            f"q<sub>\u2225</sub>={pos.x():0.3f} \u212B\u207B\u00B9, "
            f"d={2 * np.pi / np.sqrt(pos.x() ** 2 + pos.y() ** 2) * 10:0.3f} nm, "
            f"\u03B8={np.rad2deg(np.arctan2(pos.y(), pos.x())):.2f}&#176;</div>"
        )


class BetterButtons(ImageView):
    def __init__(self, *args, **kwargs):
        super(BetterButtons, self).__init__(*args, **kwargs)

        # Setup axes reset button
        self.resetAxesBtn = QPushButton("Reset Axes")
        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.resetAxesBtn.sizePolicy().hasHeightForWidth())
        self.resetAxesBtn.setSizePolicy(sizePolicy)
        self.resetAxesBtn.setObjectName("resetAxes")
        self.ui.gridLayout.addWidget(self.resetAxesBtn, 2, 1, 1, 1)
        self.resetAxesBtn.clicked.connect(self.autoRange)

        # Setup LUT reset button
        self.resetLUTBtn = QPushButton("Reset LUT")
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.resetLUTBtn.sizePolicy().hasHeightForWidth())
        # self.resetLUTBtn.setSizePolicy(sizePolicy)
        # self.resetLUTBtn.setObjectName("resetLUTBtn")
        self.ui.gridLayout.addWidget(self.resetLUTBtn, 2, 2, 1, 1)
        self.resetLUTBtn.clicked.connect(self.autoLevels)

        # Hide ROI button and rearrange
        self.ui.roiBtn.setParent(None)
        self.ui.menuBtn.setParent(None)
        # self.ui.gridLayout.addWidget(self.ui.menuBtn, 1, 1, 1, 1)
        self.ui.gridLayout.addWidget(self.ui.graphicsView, 0, 0, 2, 1)


class PolygonROI(ImageView):
    def __init__(self, *args, **kwargs):
        """
        Image view extended with an adjustable polygon region-of-interest (ROI).

        When first displayed, the polygon ROI's corners will be set to the image item's corners.

        Parameters
        ----------
        args, optional
            Positional arguments for the ImageView.
        kwargs, optional
            Keyword arguments for the ImageView.
        """
        super(PolygonROI, self).__init__(*args, **kwargs)
        rect = self.imageItem.boundingRect()  # type: QRectF
        positions = [
            (rect.bottomLeft().x(), rect.bottomLeft().y()),
            (rect.bottomRight().x(), rect.bottomRight().y()),
            (rect.topRight().x(), rect.topRight().y()),
            (rect.topLeft().x(), rect.topLeft().y()),
        ]
        self._roiItem = BetterPolyLineROI(positions=positions, closed=True, scaleSnap=True, translateSnap=True)
        self.addItem(self._roiItem)

    def __repr__(self):
        return type(self).__name__ + repr(self._roiItem)

    def poly_mask(self):
        """
        Gets the mask array for a ROI polygon on the image.

        The mask array's shape will match the image's shape.
        Any pixel inside both the ROI polygon and the image will be set to 1 in the mask array;
        all other values in the mask will be set to 0.

        Returns
        -------
        ndarray:
            Mask array of the ROI polygon within image space (mask shape matches image shape).

        """
        result, mapped = self._roiItem.getArrayRegion(
            np.ones_like(self.imageItem.image), self.imageItem, returnMappedCoords=True
        )

        # TODO -- move this code to own function and test
        # Reverse the result array to make indexing calculations easier, then revert back
        result = result[::-1, ::-1]
        mapped = mapped[::-1, ::-1]

        # Pad result mask rect into bounding rect of mask and image
        floorRow = np.floor(mapped[0]).astype(int)
        floorCol = np.floor(mapped[1]).astype(int)

        # Return empty mask if ROI bounding box does not intersect image bounding box
        resultRect = QRectF(QPointF(np.min(floorRow), np.min(floorCol)), QPointF(np.max(floorRow), np.max(floorCol)))
        if not self._intersectsImage(resultRect):
            # TODO -- is zeros(shape) the right return value for a non-intersecting polygon?
            return np.zeros(self.imageItem.image.shape)

        # Find the bounds of the ROI polygon
        minX = np.min(floorRow)
        maxX = np.max(floorRow)
        minY = np.min(floorCol)
        maxY = np.max(floorCol)

        width = self.imageItem.width()
        height = self.imageItem.height()
        # Pad the ROI polygon into the image shape
        # Don't need padding if a polygon boundary is outside of the image shape
        padXBefore = minX
        if minX < 0:
            padXBefore = 0
        padXAfter = height - maxX
        if padXAfter < 0:
            padXAfter = 0
        padYBefore = minY
        if minY < 0:
            padYBefore = 0
        padYAfter = width - maxY
        if padYAfter < 0:
            padYAfter = 0

        boundingBox = np.pad(result, ((padYBefore, padYAfter), (padXBefore, padXAfter)), "constant")

        # For trimming, any negative minimums need to be shifted into the image shape
        offsetX = 0
        offsetY = 0
        if minX < 0:
            offsetX = abs(minX)
        if minY < 0:
            offsetY = abs(minY)
        trimmed = boundingBox[abs(offsetY) : abs(offsetY) + height, abs(offsetX) : abs(offsetX) + width]

        # Reorient the trimmed mask array
        trimmed = trimmed[::-1, ::-1]

        # # TODO remove plotting code below
        # from matplotlib import pyplot as plt
        # plt.figure('bounding_box, origin="lower"')
        # plt.imshow(boundingBox, origin='lower')
        # plt.show()
        #
        #
        # plt.figure(f'trimmed, origin="lower", [{abs(offsetY)}:{abs(offsetY)+height}, {abs(offsetX)}:{abs(offsetX)+width}]')
        # plt.imshow(trimmed, origin='lower')
        # plt.show()
        # # TODO remove the plotting code above
        return trimmed

    def _intersectsImage(self, rectangle: QRectF):
        """
        Checks if a rectangle intersects the image's bounding rectangle.

        Parameters
        ----------
        rectangle
            Rectangle to test intersection with the image item's bounding rectangle.

        Returns
        -------
        bool
            True if the rectangle and the image bounding rectangle intersect; otherwise False.

        """
        # TODO -- test
        return self.imageItem.boundingRect().intersects(rectangle)


import collections
from pyqtgraph import functions as fn
from pyqtgraph import debug
from pyqtgraph import Point


class ComposableItemImageView(ImageView):
    """
    Used to compose together different image view mixins that may use different ItemImage subclasses.
    See LogScaleIntensity, LogScaleImageItem, ImageViewHistogramOverflowFIx, ImageItemHistorgramOverflowFix.
    Note that any imageItem named argument passed into the ImageView mixins above will discard the item and instead
    create a composition of imageItem_bases with their respective ImageItem class.
    """

    imageItem_bases = tuple()


class LogScaleImageItem(ImageItem):
    def __init__(self, *args, **kwargs):
        super(LogScaleImageItem, self).__init__(*args, **kwargs)
        self.logScale = True

    def render(self):
        # Convert data to QImage for display.

        profile = debug.Profiler()
        if self.image is None or self.image.size == 0:
            return
        if isinstance(self.lut, collections.Callable):
            lut = self.lut(self.image)
        else:
            lut = self.lut

        if self.logScale:
            image = self.image + 1
            with np.errstate(invalid="ignore"):
                image = image.astype(np.float)
                np.log(image, where=image >= 0, out=image)  # map to 0-255
        else:
            image = self.image

        if self.autoDownsample:
            # reduce dimensions of image based on screen resolution
            o = self.mapToDevice(QPointF(0, 0))
            x = self.mapToDevice(QPointF(1, 0))
            y = self.mapToDevice(QPointF(0, 1))
            w = Point(x - o).length()
            h = Point(y - o).length()
            if w == 0 or h == 0:
                self.qimage = None
                return
            xds = max(1, int(1.0 / w))
            yds = max(1, int(1.0 / h))
            axes = [1, 0] if self.axisOrder == "row-major" else [0, 1]
            image = fn.downsample(image, xds, axis=axes[0])
            image = fn.downsample(image, yds, axis=axes[1])
            self._lastDownsample = (xds, yds)
        else:
            pass

        # if the image data is a small int, then we can combine levels + lut
        # into a single lut for better performance
        levels = self.levels
        if levels is not None and levels.ndim == 1 and image.dtype in (np.ubyte, np.uint16):
            if self._effectiveLut is None:
                eflsize = 2 ** (image.itemsize * 8)
                ind = np.arange(eflsize)
                minlev, maxlev = levels
                levdiff = maxlev - minlev
                levdiff = 1 if levdiff == 0 else levdiff  # don't allow division by 0
                if lut is None:
                    efflut = fn.rescaleData(ind, scale=255.0 / levdiff, offset=minlev, dtype=np.ubyte)
                else:
                    lutdtype = np.min_scalar_type(lut.shape[0] - 1)
                    efflut = fn.rescaleData(
                        ind, scale=(lut.shape[0] - 1) / levdiff, offset=minlev, dtype=lutdtype, clip=(0, lut.shape[0] - 1)
                    )
                    efflut = lut[efflut]

                self._effectiveLut = efflut
            lut = self._effectiveLut
            levels = None

        # Assume images are in column-major order for backward compatibility
        # (most images are in row-major order)

        if self.axisOrder == "col-major":
            image = image.transpose((1, 0, 2)[: image.ndim])

        if self.logScale:
            with np.errstate(invalid="ignore"):
                levels = np.log(np.add(levels, 1))
            levels[0] = np.nanmax([levels[0], 0])

        argb, alpha = fn.makeARGB(image, lut=lut, levels=levels)
        self.qimage = fn.makeQImage(argb, alpha, transpose=False)


class LogScaleIntensity(ComposableItemImageView):
    def __init__(self, *args, **kwargs):
        # Composes a new type consisting of any ImageItem types in imageItem_bases with this classes's helper ImageItem
        # class (LogScaleImageItem)
        self.imageItem_bases += (LogScaleImageItem,)
        imageItem = type("DynamicImageItem", tuple(self.imageItem_bases), {})()
        if "imageItem" in kwargs:
            del kwargs["imageItem"]
        super(LogScaleIntensity, self).__init__(imageItem=imageItem, *args, **kwargs)

        self.logScale = True

        # Setup log scale button
        self.logIntensityButton = QPushButton("Log Intensity")
        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.logIntensityButton.sizePolicy().hasHeightForWidth())
        self.logIntensityButton.setSizePolicy(sizePolicy)
        self.logIntensityButton.setObjectName("logIntensity")
        self.ui.gridLayout.addWidget(self.logIntensityButton, 3, 2, 1, 1)
        self.logIntensityButton.setCheckable(True)
        self.setLogScale(True)
        self.logIntensityButton.clicked.connect(self._setLogScale)

    def _setLogScale(self, value):
        self.imageItem.logScale = value
        self.imageItem.qimage = None
        self.imageItem.update()
        self.getHistogramWidget().region.setBounds([0 if value else None, None])

    def setLogScale(self, value):
        self._setLogScale(value)
        self.logIntensityButton.setChecked(value)


class XArrayView(ImageView):
    def quickMinMax(self, data):
        """
        Estimate the min/max values of *data* by subsampling. MODIFIED TO USE THE 99TH PERCENTILE instead of max.
        """
        if data is None:
            return 0, 0

        sl = slice(None, None, max(1, int(data.size // 1e6)))
        data = np.asarray(data[sl])
        return (np.nanmin(data), np.nanpercentile(np.where(data < np.nanmax(data), data, np.nanmin(data)), 99))


class CatalogView(ImageView):
    def __init__(self, catalog=None, stream=None, field=None, *args, **kwargs):
        self.catalog = catalog
        self.stream = stream
        self.field = field
        super(CatalogView, self).__init__(*args, **kwargs)
        self.setCatalog(self.catalog, self.stream, self.field)

    def setCatalog(self, catalog, stream, field, *args, **kwargs):
        self.catalog = catalog
        self.stream = stream
        self.field = field
        if all([self.catalog, self.stream, self.field]):
            eventStream = getattr(self.catalog, self.stream).to_dask()[self.field]
            # Trim off event dimension (so transpose works)
            if eventStream.ndim > 3:
                if eventStream.shape[0] == 1:  # if only one event, drop the event axis
                    eventStream = eventStream[0]
                if eventStream.shape[1] == 1:  # if z axis is unitary, drop that axis
                    eventStream = eventStream[:, 0]
            self.xarray = MetaXArray(eventStream)
            self.setImage(img=self.xarray, *args, **kwargs)
        else:
            # TODO -- clear the current image
            pass

    def streamChanged(self, stream):
        self.setCatalog(self.catalog, stream, self.field)

    def fieldChanged(self, field):
        self.setCatalog(self.catalog, self.stream, field)
        # TODO -- figure out where to put the geometry update
        if QSpace in inspect.getmro(type(self)):
            self.setGeometry(pluginmanager.get_plugin_by_name("xicam.SAXS.calibration", "SettingsPlugin").AI(field))


class ImageItemHistogramOverflowFix(ImageItem):
    def getHistogram(self, bins="auto", step="auto", targetImageSize=200, targetHistogramSize=500, **kwds):
        """Returns x and y arrays containing the histogram values for the current image.
                For an explanation of the return format, see numpy.histogram().

                The *step* argument causes pixels to be skipped when computing the histogram to save time.
                If *step* is 'auto', then a step is chosen such that the analyzed data has
                dimensions roughly *targetImageSize* for each axis.

                The *bins* argument and any extra keyword arguments are passed to
                np.histogram(). If *bins* is 'auto', then a bin number is automatically
                chosen based on the image characteristics:

                * Integer images will have approximately *targetHistogramSize* bins,
                  with each bin having an integer width.
                * All other types will have *targetHistogramSize* bins.

                This method is also used when automatically computing levels.
                """
        if self.image is None:
            return None, None
        if step == "auto":
            step = (int(np.ceil(self.image.shape[0] / targetImageSize)), int(np.ceil(self.image.shape[1] / targetImageSize)))
        if np.isscalar(step):
            step = (step, step)
        stepData = self.image[:: step[0], :: step[1]]

        if bins == "auto":
            if stepData.dtype.kind in "ui":
                mn = stepData.min()
                mx = stepData.max()
                # print(f"\n*** mx, mn: {mx}, {mn} ({type(mx)}, {type(mn)})***\n")
                # PATCH -- explicit subtract with np.int to avoid overflow
                step = np.ceil(np.subtract(mx, mn, dtype=np.int) / 500.0)
                bins = np.arange(mn, mx + 1.01 * step, step, dtype=np.int)
                if len(bins) == 0:
                    bins = [mn, mx]
            else:
                bins = 500

        kwds["bins"] = bins
        stepData = stepData[np.isfinite(stepData)]
        hist = np.histogram(stepData, **kwds)

        return hist[1][:-1], hist[0]


class ImageViewHistogramOverflowFix(ComposableItemImageView):
    def __init__(self, *args, **kwargs):
        self.imageItem_bases += (ImageItemHistogramOverflowFix,)
        # Create a dynamic image item type, composing existing ImageItem classes with this classes's appropriate
        # ImageItem class (ImageItemHistogramOverflowFix)
        imageItem = type("DynamicImageItem", tuple(self.imageItem_bases), {})()
        if "imageItem" in kwargs:
            del kwargs["imageItem"]
        super(ImageViewHistogramOverflowFix, self).__init__(imageItem=imageItem, *args, **kwargs)



"""
This patch module adds color gradients like matplotlib's viridis etc. to PyQtGraph
"""
from pyqtgraph.graphicsItems import GradientEditorItem
from collections import OrderedDict
from qtpy.QtWidgets import QTreeWidgetItem, QWidget, QPushButton, QCheckBox
from qtpy.QtGui import QIcon
from qtpy.QtCore import Signal
from xicam.gui.static import path

from pyqtgraph.parametertree.parameterTypes import WidgetParameterItem

GradientEditorItem.__dict__["Gradients"] = OrderedDict(
    [
        (
            "thermal",
            {
                "ticks": [
                    (0.3333, (185, 0, 0, 255)),
                    (0.6666, (255, 220, 0, 255)),
                    (1, (255, 255, 255, 255)),
                    (0, (0, 0, 0, 255)),
                ],
                "mode": "rgb",
            },
        ),
        (
            "flame",
            {
                "ticks": [
                    (0.2, (7, 0, 220, 255)),
                    (0.5, (236, 0, 134, 255)),
                    (0.8, (246, 246, 0, 255)),
                    (1.0, (255, 255, 255, 255)),
                    (0.0, (0, 0, 0, 255)),
                ],
                "mode": "rgb",
            },
        ),
        (
            "yellowy",
            {
                "ticks": [
                    (0.0, (0, 0, 0, 255)),
                    (0.2328863796753704, (32, 0, 129, 255)),
                    (0.8362738179251941, (255, 255, 0, 255)),
                    (0.5257586450247, (115, 15, 255, 255)),
                    (1.0, (255, 255, 255, 255)),
                ],
                "mode": "rgb",
            },
        ),
        (
            "bipolar",
            {
                "ticks": [
                    (0.0, (0, 255, 255, 255)),
                    (1.0, (255, 255, 0, 255)),
                    (0.5, (0, 0, 0, 255)),
                    (0.25, (0, 0, 255, 255)),
                    (0.75, (255, 0, 0, 255)),
                ],
                "mode": "rgb",
            },
        ),
        (
            "viridis",
            {
                "ticks": [
                    (0.0, (68, 1, 84, 255)),
                    (0.25, (58, 82, 139, 255)),
                    (0.5, (32, 144, 140, 255)),
                    (0.75, (94, 201, 97, 255)),
                    (1.0, (253, 231, 36, 255)),
                ],
                "mode": "rgb",
            },
        ),
        (
            "inferno",
            {
                "ticks": [
                    (0.0, (0, 0, 3, 255)),
                    (0.25, (87, 15, 109, 255)),
                    (0.5, (187, 55, 84, 255)),
                    (0.75, (249, 142, 8, 255)),
                    (1.0, (252, 254, 164, 255)),
                ],
                "mode": "rgb",
            },
        ),
        (
            "plasma",
            {
                "ticks": [
                    (0.0, (12, 7, 134, 255)),
                    (0.25, (126, 3, 167, 255)),
                    (0.5, (203, 71, 119, 255)),
                    (0.75, (248, 149, 64, 255)),
                    (1.0, (239, 248, 33, 255)),
                ],
                "mode": "rgb",
            },
        ),
        (
            "magma",
            {
                "ticks": [
                    (0.0, (0, 0, 3, 255)),
                    (0.25, (80, 18, 123, 255)),
                    (0.5, (182, 54, 121, 255)),
                    (0.75, (251, 136, 97, 255)),
                    (1.0, (251, 252, 191, 255)),
                ],
                "mode": "rgb",
            },
        ),
        ("spectrum", {"ticks": [(1.0, (255, 0, 255, 255)), (0.0, (255, 0, 0, 255))], "mode": "hsv"}),
        ("cyclic", {"ticks": [(0.0, (255, 0, 4, 255)), (1.0, (255, 0, 0, 255))], "mode": "hsv"}),
        (
            "greyclip",
            {"ticks": [(0.0, (0, 0, 0, 255)), (0.99, (255, 255, 255, 255)), (1.0, (255, 0, 0, 255))], "mode": "rgb"},
        ),
        ("grey", {"ticks": [(0.0, (0, 0, 0, 255)), (1.0, (255, 255, 255, 255))], "mode": "rgb"}),
    ]
)

from pyqtgraph.parametertree import Parameter, ParameterItem, registerParameterType
from pyqtgraph.parametertree import parameterTypes
from pyqtgraph.widgets.SpinBox import SpinBox
from pyqtgraph.widgets.ColorButton import ColorButton
from pyqtgraph import ImageView, ROI, PolyLineROI
import numpy as np
from qtpy.QtCore import QSize
from qtpy.QtGui import QBrush, QPalette
import pyqtgraph


class ImageView(ImageView):
    def timeIndex(self, slider):
        ## Return the time and frame index indicated by a slider
        if self.image is None:
            return (0, 0)

        t = slider.value()

        if not hasattr(self, "tVals"):
            return (0, 0)

        xv = self.tVals
        if xv is None:
            ind = int(t)
        else:
            if len(xv) < 2:
                return (0, 0)
            inds = np.argwhere(xv <= t)  # <- The = is import to reach the last value
            if len(inds) < 1:
                return (0, t)
            ind = inds[-1, 0]
        return ind, t

    def setCurrentIndex(self, ind):
        super(ImageView, self).setCurrentIndex(ind)
        (ind, time) = self.timeIndex(self.timeLine)
        self.sigTimeChanged.emit(ind, time)


pyqtgraph.__dict__["ImageView"] = ImageView


class PolyLineROI(PolyLineROI):
    def getArrayRegion(self, data, img, axes=(0, 1), **kwds):
        """
        Return the result of ROI.getArrayRegion(), masked by the shape of the
        ROI. Values outside the ROI shape are set to 0.
        """
        br = self.boundingRect()
        if br.width() > 1000:
            raise Exception()
        sliced = ROI.getArrayRegion(self, data, img, axes=axes, fromBoundingRect=True, **kwds)
        if kwds.get("returnMappedCoords"):
            sliced, mapped = sliced

        if img.axisOrder == "col-major":
            mask = self.renderShapeMask(sliced.shape[axes[0]], sliced.shape[axes[1]])
        else:
            mask = self.renderShapeMask(sliced.shape[axes[1]], sliced.shape[axes[0]])
            mask = mask.T

        # reshape mask to ensure it is applied to the correct data axes
        shape = [1] * data.ndim
        shape[axes[0]] = sliced.shape[axes[0]]
        shape[axes[1]] = sliced.shape[axes[1]]
        mask = mask.reshape(shape)

        if kwds.get("returnMappedCoords"):
            return sliced * mask, mapped
        else:
            return sliced * mask


pyqtgraph.__dict__["PolyLineROI"] = PolyLineROI


class SafeImageView(ImageView):
    def setImage(self, img, *args, **kwargs):
        if len(getattr(img, "shape", [])) < 2:
            return
        if len(img) < 4:
            return
        super(SafeImageView, self).setImage(np.squeeze(img), *args, **kwargs)


class ImageParameterItem(WidgetParameterItem):
    def makeWidget(self):
        self.subItem = QTreeWidgetItem()
        self.addChild(self.subItem)

        w = SafeImageView()
        w.value = lambda: w.image
        w.setValue = w.setImage
        w.sigChanged = None
        # Shrink LUT
        w.getHistogramWidget().setMinimumWidth(5)
        w.ui.menuBtn.setParent(None)
        w.ui.roiBtn.setParent(None)

        self.hideWidget = False
        return w

    def treeWidgetChanged(self):
        ## TODO: fix so that superclass method can be called
        ## (WidgetParameter should just natively support this style)
        # WidgetParameterItem.treeWidgetChanged(self)
        self.treeWidget().setFirstItemColumnSpanned(self.subItem, True)
        self.treeWidget().setItemWidget(self.subItem, 0, self.widget)

        # for now, these are copied from ParameterItem.treeWidgetChanged
        self.setHidden(not self.param.opts.get("visible", True))
        self.setExpanded(self.param.opts.get("expanded", True))

    def valueChanged(self, param, val, force=False):
        ## called when the parameter's value has changed
        ParameterItem.valueChanged(self, param, val)
        if force or not np.array_equal(val, self.widget.value()):
            self.widget.setValue(val)
        self.updateDisplayLabel(val)  ## always make sure label is updated, even if values match!

    def updateDefaultBtn(self):
        pass


class ImageParameter(Parameter):
    itemClass = ImageParameterItem

    def __init__(self, *args, **kwargs):
        if "expanded" not in kwargs:
            kwargs["expanded"] = False
        super(ImageParameter, self).__init__(*args, **kwargs)

    def setValue(self, value, blockSignal=None):
        """
        Set the value of this Parameter; return the actual value that was set.
        (this may be different from the value that was requested)
        """
        try:
            if blockSignal is not None:
                self.sigValueChanged.disconnect(blockSignal)
            if np.array_equal(self.opts["value"], value):
                return value
            self.opts["value"] = value
            # self.sigValueChanged.emit(self, value)
        finally:
            if blockSignal is not None:
                self.sigValueChanged.connect(blockSignal)

        return value


registerParameterType("ndarray", ImageParameter, override=True)


class FixableWidgetParameterItem(parameterTypes.WidgetParameterItem):
    def __init__(self, param, depth):
        super(FixableWidgetParameterItem, self).__init__(param, depth)
        if param.opts.get("fixable"):
            self.fixbutton = QPushButton()
            self.fixbutton.setFixedWidth(20)
            self.fixbutton.setFixedHeight(20)
            self.fixbutton.setCheckable(True)
            self.fixbutton.setChecked(param.opts["fixed"])
            self.fixbutton.toggled.connect(param.sigFixToggled)
            self.fixbutton.toggled.connect(lambda fixed: param.setOpts(fixed=fixed))
            # self.fixbutton.toggled.connect(lambda fixed: self.widgetValueChanged())

            self.fixbutton.setIcon(QIcon(path("icons/anchor.png")))
            self.layoutWidget.layout().addWidget(self.fixbutton)

    def optsChanged(self, param, opts):
        """Called when any options are changed that are not
        name, value, default, or limits"""
        # print "opts changed:", opts
        ParameterItem.optsChanged(self, param, opts)

        if "readonly" in opts:
            self.updateDefaultBtn()
            if isinstance(self.widget, (QCheckBox, ColorButton)):
                self.widget.setEnabled(not opts["readonly"])

        ## If widget is a SpinBox, pass options straight through
        if isinstance(self.widget, SpinBox):
            if "units" in opts and "suffix" not in opts:
                opts["suffix"] = opts["units"]
            try:  # patch passes silently for 'fixed'
                self.widget.setOpts(**opts)
            except TypeError:
                pass
            self.updateDisplayLabel()


class FixableSimpleParameter(parameterTypes.SimpleParameter):
    itemClass = FixableWidgetParameterItem
    sigFixToggled = Signal(bool)


registerParameterType("int", FixableSimpleParameter, override=True)
registerParameterType("float", FixableSimpleParameter, override=True)


class BetterGroupParameteritem(parameterTypes.GroupParameterItem):
    def updateDepth(self, depth):

        ## Change item's appearance based on its depth in the tree
        ## This allows highest-level groups to be displayed more prominently.
        if depth == 0:
            for c in [0, 1]:
                self.setBackground(c, QBrush(QPalette().color(QPalette.Light)))
                # self.setForeground(c, QBrush(QPalette().color(QPalette.Dark)))
                font = self.font(c)
                font.setBold(True)
                font.setPointSize(font.pointSize() + 1)
                self.setFont(c, font)
                self.setSizeHint(0, QSize(0, 25))
        else:
            for c in [0, 1]:
                self.setBackground(c, QBrush(QPalette().color(QPalette.Light)))
                font = self.font(c)
                font.setBold(True)
                # font.setPointSize(font.pointSize()+1)
                self.setFont(c, font)
                self.setSizeHint(0, QSize(0, 20))


class BetterGroupParameter(parameterTypes.GroupParameter):
    itemClass = BetterGroupParameteritem


parameterTypes.GroupParameter = BetterGroupParameter
registerParameterType("group", BetterGroupParameter, override=True)


class CounterGroupParameterItem(BetterGroupParameteritem):
    def __init__(self, *args, **kwargs):
        super(CounterGroupParameterItem, self).__init__(*args, **kwargs)
        self.param.sigChildAdded.connect(self.updateText)
        self.param.sigChildRemoved.connect(self.updateText)

    def updateText(self, *_):
        self.setText(0, f"{self.param.name()} ({self.childCount()})")


class CounterGroupParameter(BetterGroupParameter):
    itemClass = CounterGroupParameterItem


registerParameterType("countergroup", CounterGroupParameter, override=True)


class LazyGroupParameterItem(BetterGroupParameteritem):
    # Note: no accounting for if items are removed

    def __init__(self, *args, **kwargs):
        super(LazyGroupParameterItem, self).__init__(*args, **kwargs)
        self._queuedchildren = []

    @staticmethod
    def initialize_treewidget(treewidget):
        treewidget.itemExpanded.connect(LazyGroupParameterItem.itemExpanded)

    @staticmethod
    def itemExpanded(item):
        for i in range(item.childCount()):
            if hasattr(item.child(i), "flushChildren"):
                item.child(i).flushChildren()

    def childAdded(self, *childargs):
        self._queuedchildren.append(childargs)
        if self.isExpanded():
            self.flushChildren()

    def flushChildren(self):
        while len(self._queuedchildren):
            super(LazyGroupParameterItem, self).childAdded(*self._queuedchildren.pop())


class LazyGroupParameter(BetterGroupParameter):
    itemClass = LazyGroupParameterItem


registerParameterType("lazygroup", LazyGroupParameter, override=True)



import pickle
from qtpy.QtCore import QAbstractTableModel, QMimeData, Qt, Signal, QSize
from qtpy.QtGui import QIcon, QPixmap
from qtpy.QtWidgets import QSplitter, QApplication, QWidget, QAbstractItemView, QToolBar, QToolButton, QMenu, \
    QVBoxLayout, QTableView, QItemDelegate, QGridLayout, QLabel, QPushButton, QSizePolicy, QHeaderView, QCheckBox, \
    QHBoxLayout
from xicam.core.execution.workflow import Workflow
from xicam.plugins import OperationPlugin
from pyqtgraph.parametertree import ParameterTree, Parameter
from pyqtgraph.parametertree.parameterTypes import GroupParameter
from xicam.gui.static import path
from xicam.plugins import manager as pluginmanager
from functools import partial
from typing import List
from xicam.plugins import manager as pluginmanager, OperationPlugin
from functools import partial, lru_cache


# WorkflowEditor
#  WorkflowOperationEditor
# WorkflowWidget
#  LinearWorkflowView
#   WorkflowModel


class WorkflowEditor(QSplitter):
    sigWorkflowChanged = Signal()
    sigRunWorkflow = Signal()

    def __init__(self, workflow: Workflow, **kwargs):
        super(WorkflowEditor, self).__init__()
        self.workflow = workflow
        self.kwargs = kwargs
        self.setOrientation(Qt.Vertical)

        self.operationeditor = WorkflowOperationEditor()
        self.workflowview = LinearWorkflowView(WorkflowModel(workflow))

        self.addWidget(self.operationeditor)
        workflow_widget = WorkflowWidget(self.workflowview)
        self.addWidget(workflow_widget)
        workflow_widget.sigRunWorkflow.connect(self.sigRunWorkflow.emit)
        # Should this work internally? How would the start operations get their inputs?
        # Would the ExamplePlugin need to explicitly set the parameter value (even for hidden image)?
        # It would be nice to just have this work easily... (to ExamplePlugin's perspective)
        # workflow_widget.sigRunWorkflow.connect(self.run_workflow)
        # TODO make work for autorun...
        # OR is this the outside class's respsonsibility (see SAXSGUIPlugin.maskeditor)

        self.workflowview.sigShowParameter.connect(self.setParameters)

        workflow.attach(self.sigWorkflowChanged.emit)

    def run_workflow(self, _):
        self.workflow.execute(**self.kwargs)

    def setParameters(self, operation: OperationPlugin):
        if operation:
            # Create a new Parameter from the emitted operation,
            # Then wire up its connections for use in a parameter tree.
            parameter = operation.as_parameter()
            group = GroupParameter(name='Selected Operation', children=parameter)
            operation.wireup_parameter(group)

            # Add the Parameter to the parameter tree
            group.blockSignals(True)
            for child in group.children():
                child.blockSignals(True)
            self.operationeditor.setParameters(group, showTop=False)
            QApplication.processEvents()
            group.blockSignals(False)
            for child in group.children():
                child.blockSignals(False)


class WorkflowOperationEditor(ParameterTree):
    pass


class WorkflowWidget(QWidget):
    sigAddFunction = Signal(object)
    sigRunWorkflow = Signal(object)
    # TODO -- emit Workflow from sigRunWorkflow

    def __init__(self, workflowview: QAbstractItemView):
        super(WorkflowWidget, self).__init__()

        self.view = workflowview

        self.autorun_checkbox = QCheckBox("Run Automatically")
        self.autorun_checkbox.setCheckState(Qt.Unchecked)
        self.autorun_checkbox.stateChanged.connect(self._autorun_state_changed)
        self.run_button = QPushButton("Run Workflow")
        self.run_button.clicked.connect(self.sigRunWorkflow.emit)
        # TODO -- actually hook up the auto run OR dependent class needs to connect (see SAXSGUIPlugin)

        self.toolbar = QToolBar()
        self.addfunctionmenu = QToolButton()
        self.addfunctionmenu.setIcon(QIcon(path("icons/addfunction.png")))
        self.addfunctionmenu.setText("Add Function")
        # Defer menu population to once the plugins have been loaded; otherwise, the menu may not contain anything
        # if this widget is init'd before all plugins have been loaded.
        self.functionmenu = QMenu()
        self.functionmenu.aboutToShow.connect(self.populateFunctionMenu)
        self.addfunctionmenu.setMenu(self.functionmenu)
        self.addfunctionmenu.setPopupMode(QToolButton.InstantPopup)
        self.toolbar.addWidget(self.addfunctionmenu)
        # self.toolbar.addAction(QIcon(path('icons/up.png')), 'Move Up')
        # self.toolbar.addAction(QIcon(path('icons/down.png')), 'Move Down')
        self.toolbar.addAction(QIcon(path("icons/folder.png")), "Load Workflow")
        self.toolbar.addAction(QIcon(path("icons/trash.png")), "Delete Operation", self.deleteOperation)

        v = QVBoxLayout()
        v.addWidget(self.view)
        h = QHBoxLayout()
        h.addWidget(self.autorun_checkbox)
        h.addWidget(self.run_button)
        v.addLayout(h)
        v.addWidget(self.toolbar)
        v.setContentsMargins(0, 0, 0, 0)
        self.setLayout(v)

    def _autorun_state_changed(self, state):
        if state == Qt.Checked:
            self.run_button.setDisabled(True)
        else:
            self.run_button.setDisabled(False)

    def _run_workflow(self, _):
        self._workflow

    def populateFunctionMenu(self):
        self.functionmenu.clear()
        sortingDict = {}
        for plugin in pluginmanager.get_plugins_of_type("OperationPlugin"):
            typeOfOperationPlugin = plugin.categories
            # TODO : should OperationPlugin be responsible for initializing categories
            # to some placeholder value (instead of [])?
            if typeOfOperationPlugin == []:
                typeOfOperationPlugin = "uncategorized"  # put found operations into a default category
            if not typeOfOperationPlugin in sortingDict.keys():
                sortingDict[typeOfOperationPlugin] = []
            sortingDict[typeOfOperationPlugin].append(plugin)
        for key in sortingDict.keys():
            self.functionmenu.addSeparator()
            self.functionmenu.addAction(key)
            self.functionmenu.addSeparator()
            for plugin in sortingDict[key]:
                self.functionmenu.addAction(plugin.name, partial(self.addOperation, plugin, autoconnectall=True))

    def addOperation(self, operation: OperationPlugin, autoconnectall=True):
        self.view.model().workflow.add_operation(operation())
        if autoconnectall:
            self.view.model().workflow.auto_connect_all()
        print("selected new row:", self.view.model().rowCount() - 1)
        self.view.setCurrentIndex(self.view.model().index(self.view.model().rowCount() - 1, 0))

    def deleteOperation(self):
        for index in self.view.selectedIndexes():
            operation = self.view.model().workflow.operations[index.row()]
            self.view.model().workflow.remove_operation(operation)


class LinearWorkflowView(QTableView):
    sigShowParameter = Signal(object)

    def __init__(self, workflowmodel=None, *args, **kwargs):
        super(LinearWorkflowView, self).__init__(*args, **kwargs)

        self.setItemDelegateForColumn(0, DisableDelegate(self))
        self.setItemDelegateForColumn(1, HintsDelegate(self))

        self.setModel(workflowmodel)
        workflowmodel.workflow.attach(self.selectionChanged)

        self.horizontalHeader().close()
        # self.horizontalHeader().setStretchLastSection(True)
        self.horizontalHeader().setResizeMode(QHeaderView.ResizeToContents)
        self.horizontalHeader().setResizeMode(1, QHeaderView.Stretch)

        # self.horizontalHeader().setSectionMovable(True)
        # self.horizontalHeader().setDragEnabled(True)
        # self.horizontalHeader().setDragDropMode(QAbstractItemView.InternalMove)
        self.verticalHeader().setResizeMode(QHeaderView.ResizeToContents)
        self.setDragDropMode(QAbstractItemView.InternalMove)
        self.setDragEnabled(True)
        self.setDropIndicatorShown(True)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)

    def selectionChanged(self, selected=None, deselected=None):
        if self.selectedIndexes() and self.selectedIndexes()[0].row() < self.model().rowCount():
            operation = self.model().workflow.operations[self.selectedIndexes()[0].row()]  # type: OperationPlugin
            self.sigShowParameter.emit(operation)
        else:
            self.sigShowParameter.emit(None)
        for child in self.children():
            if hasattr(child, "repaint"):
                child.repaint()

        selectedrows = set(map(lambda index: index.row(), self.selectedIndexes()))
        for row in range(self.model().rowCount()):
            widget = self.indexWidget(self.model().index(row, 1))
            if hasattr(widget, "setSelectedVisibility"):
                widget.setSelectedVisibility(row in selectedrows)
        # self.resizeRowsToContents()


class WorkflowModel(QAbstractTableModel):
    def __init__(self, workflow: Workflow):
        self.workflow = workflow
        super(WorkflowModel, self).__init__()

        self.workflow.attach(partial(self.layoutChanged.emit))

    def mimeTypes(self):
        return ["text/plain"]

    def mimeData(self, indexes):
        mimedata = QMimeData()
        mimedata.setText(str(indexes[0].row()))
        return mimedata

    def dropMimeData(self, data, action, row, column, parent):
        srcindex = int(data.text())
        operation = self.workflow.operations[srcindex]
        self.workflow.remove_operation(operation)
        self.workflow.insert_operation(parent.row(), operation)
        self.workflow.auto_connect_all()
        return True

    def supportedDropActions(self):
        return Qt.CopyAction | Qt.MoveAction

    def flags(self, index):
        if not index.isValid():
            return Qt.ItemIsEnabled
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled | Qt.ItemIsDropEnabled

    def rowCount(self, *args, **kwargs):
        return len(self.workflow.operations)

    def columnCount(self, *args, **kwargs):
        return 2

    def data(self, index, role):
        operation = self.workflow.operations[index.row()]
        if not index.isValid():
            return None
        elif role != Qt.DisplayRole:
            return None
        elif index.column() == 0:
            return partial(self.workflow.toggle_disabled, operation)
        elif index.column() == 1:
            # return getattr(process, 'name', process.__class__.__name__)
            return None
        return ""

    def headerData(self, col, orientation, role):
        return None


class HintsDelegate(QItemDelegate):
    def __init__(self, parent):
        super(HintsDelegate, self).__init__(parent=parent)
        self.view = parent

    def paint(self, painter, option, index):
        if not (self.view.indexWidget(index)):
            # selected = index in map(lambda index: index.row, self.view.selectedIndexes())
            operation = self.view.model().workflow.operations[index.row()]
            widget = HintsWidget(operation, self.view, index)
            self.view.setIndexWidget(index, widget)


class HintsWidget(QWidget):
    def __init__(self, operation, view, index):
        super(HintsWidget, self).__init__()
        self.view = view
        self.setLayout(QGridLayout())
        self.layout().addWidget(QLabel(operation.name), 0, 0, 1, 2)
        self.hints = operation.hints

        enabledhints = [hint for hint in self.hints if hint.enabled]

        for i, hint in enumerate(enabledhints):
            enablebutton = QPushButton(icon=mk_enableicon())
            sp = QSizePolicy()
            sp.setWidthForHeight(True)
            enablebutton.setSizePolicy(sp)
            enablebutton.setVisible(False)
            label = QLabel(hint.name)
            label.setVisible(False)
            self.layout().addWidget(enablebutton, i + 1, 0, 1, 1)
            self.layout().addWidget(label)

        self.name = operation.name

        print('size1:', operation.name, self.sizeHint())

    def setSelectedVisibility(self, selected):
        for row in range(1, self.layout().rowCount()):
            self.layout().itemAtPosition(row, 0).widget().setVisible(selected)
            self.layout().itemAtPosition(row, 1).widget().setVisible(selected)
        print('size2:', self.name, self.sizeHint())

    def sizeHint(self):
        return QSize(30, 30)


class DeleteDelegate(QItemDelegate):
    def __init__(self, parent=None):
        super(DeleteDelegate, self).__init__(parent)
        self._parent = parent

    def paint(self, painter, option, index):
        if not self._parent.indexWidget(index):
            button = QToolButton(self.parent())
            button.setAutoRaise(True)
            button.setText("Delete Operation")
            button.setIcon(QIcon(path("icons/trash.png")))
            sp = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
            sp.setWidthForHeight(True)
            button.setSizePolicy(sp)
            button.clicked.connect(index.data())

            self._parent.setIndexWidget(index, button)


class DisableDelegate(QItemDelegate):
    class DelegateClass(QToolButton):
        def __init__(self, parent=None):
            super(DisableDelegate.DelegateClass, self).__init__(parent=parent)
            self.setText("i")
            self.setAutoRaise(True)

            self.setIcon(mk_enableicon())
            self.setCheckable(True)
            sp = QSizePolicy()
            sp.setWidthForHeight(True)
            self.setSizePolicy(sp)

    def __init__(self, parent):
        super(DisableDelegate, self).__init__(parent)
        self._parent = parent

    def paint(self, painter, option, index):
        if not self._parent.indexWidget(index):
            button = self.DelegateClass(self.parent())
            button.clicked.connect(index.data())
            self._parent.setIndexWidget(index, button)

    def sizeHint(self, QStyleOptionViewItem, QModelIndex):
        return QSize(30, 30)


@lru_cache()
def mk_enableicon():
    enableicon = QIcon()
    enableicon.addPixmap(QPixmap(path("icons/enable.png")), state=enableicon.Off)
    enableicon.addPixmap(QPixmap(path("icons/disable.png")), state=enableicon.On)
    return enableicon



from qtpy.QtCore import *
from qtpy.QtWidgets import *
from .plugin import PluginType


# TODO: make classes usable without qt


# class _metaQWidgetPlugin(type(QObject), type(IPlugin)):
#     pass
#
#
# class QWidgetPlugin(metaclass=_metaQWidgetPlugin):
#     isSingleton = False


class QWidgetPlugin(QWidget, PluginType):
    is_singleton = False
    needs_qt = True


def test_QWidgetPlugin():
    from pyqtgraph import ImageView

    class ImageViewPlugin(QWidgetPlugin, ImageView):
        pass

    app = makeapp()
    i = ImageViewPlugin()
    i.show()
    t = QTimer()
    t.singleShot(1000, i.close)
    mainloop()


def makeapp():
    app = QApplication([])
    return app


def mainloop():
    app = QApplication.instance()
    app.exec_()



import pyqtgraph as pg
from qtpy.QtWidgets import QTabWidget
from typing import List


class TabPlotWidget(QTabWidget):
    def __init__(self):
        super(TabPlotWidget, self).__init__()

    def plot(self, *args, **kwargs):
        plotwidget = pg.PlotWidget(*args, **kwargs)
        self.addTab(kwargs.get("title", "Untitled"), plotwidget)

    @property
    def plots(self) -> List[pg.PlotWidget]:
        return [self.widget(i) for i in range(self.count())]



import platform
import os
import os.path
import sys
import venv

import pathlib
from appdirs import user_config_dir, site_config_dir, user_cache_dir
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QLabel

from xicam.gui.static import path
from xicam.plugins import SettingsPlugin

op_sys = platform.system()
if op_sys == "Darwin":  # User config dir incompatible with venv on darwin (space in path name conflicts)
    user_venv_dir = os.path.join(user_cache_dir(appname="xicam"), "venvs")
else:
    user_venv_dir = os.path.join(user_config_dir(appname="xicam"), "venvs")
site_venv_dir = os.path.join(site_config_dir(appname="xicam"), "venvs")

venvs = {}
observers = []

# Python 2 style execfile function
execfile = lambda filename, globals=None, locals=None: exec(open(filename).read(), globals, locals)


# TODO: transition to http://virtualenvwrapper.readthedocs.io/en/latest/index.html


class VenvsSettingsPlugin(SettingsPlugin):
    def __init__(self):
        if not current_environment:
            initialize_venv()

        self.widget = QLabel("test")
        super(VenvsSettingsPlugin, self).__init__(QIcon(str(path("icons/python.png"))), "Virtual Environments", self.widget)

    def toState(self):
        return None  # self.parameter.saveState()

    def fromState(self, state):
        pass  # self.parameter.restoreState(state)


def create_environment(name: str):
    """
    Create a new sandbox environment in the user_venv_dir with name name.

    Parameters
    ----------
    name : str
        Name of sandbox environment to create.
    """

    envpath = pathlib.Path(user_venv_dir, name)
    if envpath.is_dir():
        return
        # raise ValueError('Environment already exists.')
    venv.EnvBuilder(with_pip=False).create(envpath)


def use_environment(name):
    """
    Activate the sandbox environment with name name in user_venv_dir

    Parameters
    ----------
    name : str
        Name of sandbox environment to activate
    """
    path = pathlib.Path(user_venv_dir, name)
    if not path.is_dir():
        raise ValueError(f"Sandbox environment '{name}' could not be found.")

    global current_environment
    current_environment = str(path)
    activate_this(current_environment)

    for observer in observers:
        observer.venvChanged()


def activate_this(path):
    # Below copied and modified from the activate_this.py module of virtualenv, which is missing form venv
    old_os_path = os.environ.get("PATH", "")
    os.environ["PATH"] = os.path.dirname(os.path.abspath(path)) + os.pathsep + old_os_path
    if sys.platform == "win32":
        site_packages = os.path.join(path, "Lib", "site-packages")
    else:
        site_packages = os.path.join(path, "lib", "python%s" % sys.version[:3], "site-packages")
    prev_sys_path = list(sys.path)
    # if not getattr(sys, 'frozen', False):  # site missing addsitedir when frozen
    try:
        import site

        addsitedir = site.addsitedir
    except AttributeError:  # frozen apps have no site-packages; and thus their site.py is fake
        # NOTE: relevant methods have been extracted from a real site.py
        def makepath(*paths):
            dir = os.path.join(*paths)
            try:
                dir = os.path.abspath(dir)
            except OSError:
                pass
            return dir, os.path.normcase(dir)

        def _init_pathinfo():
            """Return a set containing all existing file system items from sys.path."""
            d = set()
            for item in sys.path:
                try:
                    if os.path.exists(item):
                        _, itemcase = makepath(item)
                        d.add(itemcase)
                except TypeError:
                    continue
            return d

        def addpackage(sitedir, name, known_paths):
            """Process a .pth file within the site-packages directory:
               For each line in the file, either combine it with sitedir to a path
               and add that to known_paths, or execute it if it starts with 'import '.
            """
            if known_paths is None:
                known_paths = _init_pathinfo()
                reset = True
            else:
                reset = False
            fullname = os.path.join(sitedir, name)
            try:
                f = open(fullname, "r")
            except OSError:
                return
            with f:
                for n, line in enumerate(f):
                    if line.startswith("#"):
                        continue
                    try:
                        if line.startswith(("import ", "import\t")):
                            exec(line)
                            continue
                        line = line.rstrip()
                        dir, dircase = makepath(sitedir, line)
                        if not dircase in known_paths and os.path.exists(dir):
                            sys.path.append(dir)
                            known_paths.add(dircase)
                    except Exception:
                        print("Error processing line {:d} of {}:\n".format(n + 1, fullname), file=sys.stderr)
                        import traceback

                        for record in traceback.format_exception(*sys.exc_info()):
                            for line in record.splitlines():
                                print("  " + line, file=sys.stderr)
                        print("\nRemainder of file ignored", file=sys.stderr)
                        break
            if reset:
                known_paths = None
            return known_paths

        def addsitedir(sitedir, known_paths=None):
            """Add 'sitedir' argument to sys.path if missing and handle .pth files in
            'sitedir'"""
            if known_paths is None:
                known_paths = _init_pathinfo()
                reset = True
            else:
                reset = False
            sitedir, sitedircase = makepath(sitedir)
            if not sitedircase in known_paths:
                sys.path.append(sitedir)  # Add path component
                known_paths.add(sitedircase)
            try:
                names = os.listdir(sitedir)
            except OSError:
                return
            names = [name for name in names if name.endswith(".pth")]
            for name in sorted(names):
                addpackage(sitedir, name, known_paths)
            if reset:
                known_paths = None
            return known_paths

    addsitedir(site_packages)

    sys.real_prefix = sys.prefix
    sys.prefix = path
    # Move the added items to the front of the path:
    new_sys_path = []
    for item in list(sys.path):
        if item not in prev_sys_path:
            new_sys_path.append(item)
            sys.path.remove(item)
    sys.path[:0] = new_sys_path


current_environment = ""


# TODO: find all venvs; populate the venvs global
def initialize_venv():
    global current_environment
    create_environment("default")
    use_environment("default")
    current_environment = str(pathlib.Path(user_venv_dir, "default"))



import sys
import os
import signal
import trace
from xicam.core.args import parse_args

print("args:", sys.argv)
print("path:", sys.path)

if sys.argv[0].endswith("Xi-cam"):
    root = os.path.dirname(sys.argv[0])
    sys.path = [path for path in sys.path if os.path.abspath(root) in os.path.abspath(path)]

# Quickly extract zip file to make imports easier
if ".zip/" in os.__file__:
    import zipfile

    zip_ref = zipfile.ZipFile(os.path.dirname(os.__file__), "r")
    zip_ref.extractall(os.path.dirname(os.path.dirname(os.__file__)))
    zip_ref.close()

os.environ["QT_API"] = "pyqt5"
import qtpy
from qtpy.QtWidgets import QApplication, QErrorMessage
from qtpy.QtCore import QCoreApplication, QProcess

if qtpy.API_NAME == "PyQt5" and "PySide" in sys.modules:
    del sys.modules["PySide"]

QCoreApplication.setOrganizationName("Camera")
QCoreApplication.setApplicationName("Xi-cam")


def _main(args, exec=True):
    # import pydm
    # app = QApplication([])
    # app = pydm.PyDMApplication()
    app = QApplication.instance() or QApplication([])
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    from xicam.gui.windows import splash
    from xicam.core import msg

    if getattr(args, 'verbose', False):
        QErrorMessage.qtHandler()

    # start splash in subprocess
    splash_proc = QProcess()
    # splash_proc.started.connect(lambda: print('started splash'))
    # splash_proc.finished.connect(lambda: print('finished splashing'))
    log_file = msg.file_handler.baseFilename
    initial_length = os.path.getsize(log_file)
    splash_proc.start(sys.executable, [splash.__file__, log_file, str(initial_length)])

    from xicam.gui.windows.mainwindow import XicamMainWindow

    mainwindow = XicamMainWindow()
    while splash_proc.state() != QProcess.NotRunning:
        app.processEvents()
    # splash_proc.waitForFinished()
    mainwindow.show()
    mainwindow.activateWindow()

    # splash = splash.XicamSplashScreen(args=args)
    if exec:
        return app.exec_()
    else:
        return mainwindow


def main():
    args = parse_args(exit_on_fail=True)
    if args.verbose > 1:
        tracer = trace.Trace(count=False, trace=True)
        tracer.run("_main(args)")
    else:
        return _main(args)


if __name__ == "__main__":
    return_code = main()
    sys.exit(return_code)
# TODO: check entry log when running entry point



from qtpy.QtWidgets import QApplication
import pyqtgraph as pg
import numpy as np


class HoverHighlight(pg.PlotWidget):
    """
    Highlights any scatter spots moused-over, giving a feel that they can be clicked on for more info
    """

    def __init__(self, *args, **kwargs):
        super(HoverHighlight, self).__init__(*args, **kwargs)
        self._last_highlighted = None
        self._last_pen = None

    def mouseMoveEvent(self, ev):
        if self._last_highlighted:
            self._last_highlighted.setPen(self._last_pen)
            self._last_highlighted = None

        super(HoverHighlight, self).mouseMoveEvent(ev)

        if self.plotItem.boundingRect().contains(ev.pos()):
            mousePoint = self.plotItem.mapToView(pg.Point(ev.pos()))

            for item in self.scene().items():
                if isinstance(item, pg.PlotDataItem):
                    if item.curve.mouseShape().contains(mousePoint):
                        scatter = item.scatter  # type: pg.ScatterPlotItem
                        points = scatter.pointsAt(mousePoint)
                        if points:
                            self._last_pen = points[0].pen()
                            self._last_highlighted = points[0]
                            points[0].setPen(pg.mkPen("w", width=2))
                            break


class CurveLabels(HoverHighlight):
    def __init__(self, *args, **kwargs):
        super(CurveLabels, self).__init__(*args, **kwargs)

        self._arrow = None
        self._text = None
        self._curvepoint = None

    def plot(self, *args, **kwargs):
        if "symbolSize" not in kwargs:
            kwargs["symbolSize"] = 10
        if "symbol" not in kwargs:
            kwargs["symbol"] = "o"
        if "symbolPen" not in kwargs:
            kwargs["symbolPen"] = pg.mkPen((0, 0, 0, 0))
        if "symbolBrush" not in kwargs:
            kwargs["symbolBrush"] = pg.mkBrush((0, 0, 0, 0))

        item = self.plotItem.plot(*args, **kwargs)

        item.sigPointsClicked.connect(self.showLabel)
        return item

    def showLabel(self, item, points):
        if self._curvepoint:
            self.scene().removeItem(self._arrow)
            self.scene().removeItem(self._text)

        point = points[0]
        self._curvepoint = pg.CurvePoint(item.curve)
        self.addItem(self._curvepoint)
        self._arrow = pg.ArrowItem(angle=90)
        self._arrow.setParentItem(self._curvepoint)
        self._arrow.setZValue(10000)
        self._text = pg.TextItem(item.name(), anchor=(0.5, -1.0), border=pg.mkPen("w"), fill=pg.mkBrush("k"))
        self._text.setParentItem(self._curvepoint)

        self._curvepoint.setIndex(list(item.scatter.points()).index(point))


if __name__ == "__main__":
    qapp = QApplication([])
    w = CurveLabels()
    for i in range(10):
        pen = pg.mkColor((i, 10))
        w.plot(np.random.random((100,)) + i * 0.5, name=str(i), pen=pen)

    w.show()

    qapp.exec_()



from qtpy.QtCore import *
from qtpy.QtWidgets import *
from .plugin import PluginType


class ControllerPlugin(QWidget, PluginType):
    needs_qt = True
    is_singleton = False

    def __init__(self, device, parent=None, *args):
        self.device = device
        PluginType.__init__(self)
        QWidget.__init__(
            self, parent, *args
        )  # TODO: Strange MRO, not sure why QWidget init isn't being called; replaced by IPlugin init twice?



from pyqtgraph import ImageView
import numpy as np


class DynImageView(ImageView):
    def __init__(self, *args, **kwargs):
        super(DynImageView, self).__init__(*args, **kwargs)

        # Use Viridis by default
        self.setPredefinedGradient("viridis")

        # Shrink LUT
        self.getHistogramWidget().setMinimumWidth(10)

        # Don't invert Y axis
        self.view.invertY(False)
        self.imageItem.setOpts(axisOrder="row-major")

        # Setup late signal
        self.sigTimeChangeFinished = self.timeLine.sigPositionChangeFinished

    def quickMinMax(self, data):
        """
        Estimate the min/max values of *data* by subsampling. MODIFIED TO USE THE 99TH PERCENTILE instead of max.
        """
        if data is None:
            return 0, 0
        ax = np.argmax(data.shape)
        sl = [slice(None)] * data.ndim
        sl[ax] = slice(None, None, max(1, int(data.size // 1e6)))
        data = data[sl]
        return (
            np.nanpercentile(np.where(data > np.nanmin(data), data, np.nanmax(data)), 2),
            np.nanpercentile(np.where(data < np.nanmax(data), data, np.nanmin(data)), 98),
        )

    def setImage(self, img, **kwargs):
        super(DynImageView, self).setImage(img, **kwargs)
        if len(img.shape) > 2 and img.shape[0] == 1:
            self.ui.roiPlot.hide()



from functools import partial

from intake.catalog import Catalog
from intake.catalog.entry import CatalogEntry
from qtpy.QtCore import QPropertyAnimation, QPoint, QEasingCurve, Qt, Slot, Signal, QSettings, QUrl
from qtpy.QtGui import QDesktopServices, QIcon, QPixmap, QKeySequence, QFont
from qtpy.QtWidgets import (
    QMainWindow,
    QApplication,
    QStatusBar,
    QProgressBar,
    QStackedWidget,
    QMenu,
    QShortcut,
    QDockWidget,
    QWidget,
    QToolBar,
    QActionGroup,
    QGraphicsOpacityEffect,
    QAction,
    QSpinBox,
    QMessageBox,
)
from xicam import _version as version

from xicam.plugins import manager as pluginmanager
from xicam.plugins import PluginType
from xicam.plugins.guiplugin import PanelState
from xicam.gui.widgets.debugmenubar import DebuggableMenuBar
from xicam.core import msg, threads
from xicam.core.data import NonDBHeader
from xicam.gui.settings.appearance import AppearanceSettingsPlugin
from ..widgets import get_default_stage
from .settings import ConfigDialog, ConfigRestorer
from ..static import path


class XicamMainWindow(QMainWindow):
    """
    The Xi-cam main window. Includes layout for various panels and mechanism to position GUIPlugin contents into panels.
    """

    def __init__(self):
        super(XicamMainWindow, self).__init__()

        # Set icon
        self.setWindowIcon(QIcon(QPixmap(str(path("icons/xicam.gif")))))

        # Set size and position
        self.setGeometry(0, 0, 1000, 600)
        frameGm = self.frameGeometry()
        screen = QApplication.desktop().screenNumber(QApplication.desktop().cursor().pos())
        centerPoint = QApplication.desktop().screenGeometry(screen).center()
        frameGm.moveCenter(centerPoint)
        self.move(frameGm.topLeft())

        # Init child widgets to None
        self.topwidget = (
            self.leftwidget
        ) = (
            self.rightwidget
        ) = (
            self.bottomwidget
        ) = self.lefttopwidget = self.righttopwidget = self.leftbottomwidget = self.rightbottomwidget = None

        # Setup appearance
        self.setWindowTitle("Xi-cam")
        self.load_style()

        # Attach an object to restore config when loaded
        self._config_restorer = ConfigRestorer()

        # Load plugins
        pluginmanager.qt_is_safe = True
        pluginmanager.initialize_types()
        pluginmanager.collect_plugins()

        # Setup center/toolbar/statusbar/progressbar
        self.pluginmodewidget = pluginModeWidget()
        self.pluginmodewidget.sigSetStage.connect(self.setStage)
        self.pluginmodewidget.sigSetGUIPlugin.connect(self.setGUIPlugin)
        self.addToolBar(self.pluginmodewidget)
        self.setStatusBar(QStatusBar(self))
        msg.progressbar = QProgressBar(self)
        msg.progressbar.hide()
        msg.statusbar = self.statusBar()
        self.statusBar().addPermanentWidget(msg.progressbar)
        self.setCentralWidget(QStackedWidget())
        # NOTE: CentralWidgets are force-deleted when replaced, even if the object is still referenced;
        # To avoid this, a QStackedWidget is used for the central widget.

        # Setup menubar
        menubar = DebuggableMenuBar()
        self.setMenuBar(menubar)
        file = QMenu("&File", parent=menubar)
        menubar.addMenu(file)
        file.addAction("Se&ttings", self.showSettings, shortcut=QKeySequence(Qt.CTRL + Qt.ALT + Qt.Key_S))
        file.addAction("E&xit", self.close)

        # Set up help
        help = QMenu("&Help", parent=menubar)
        documentation_link = QUrl("https://xi-cam.readthedocs.io/en/latest/")
        help.addAction("Xi-CAM &Help", lambda: QDesktopServices.openUrl(documentation_link))
        slack_link = QUrl("https://nikea.slack.com")
        help.addAction("Chat on &Slack", lambda: QDesktopServices.openUrl(slack_link))
        help.addSeparator()

        about_title = "About Xi-CAM"
        version_text = f"""Version: <strong>{version.get_versions()['version']}</strong>"""
        copyright_text = f"""<small>Copyright (c) 2016, The Regents of the University of California, \
            through Lawrence Berkeley National Laboratory \
            (subject to receipt of any required approvals from the U.S. Dept. of Energy). \
            All rights reserved.</small>"""
        funding_text = f"""Funding for this research was provided by: \
            Lawrence Berkeley National Laboratory (grant No. TReXS LDRD to AH); \
            US Department of Energy (award No. Early Career Award to AH; \
            contract No. DE-SC0012704; contract No. DE-AC02-06CH11357; \
            contract No. DE-AC02-76SF00515; contract No. DE-AC02-05CH11231); \
            Center for Advanced Mathematics in Energy Research Applications; \
            Light Source Directors Data Solution Task Force Pilot Project."""
        about_text = version_text + "<br><br>" + funding_text + "<br><hr>" + copyright_text
        about_box = QMessageBox(QMessageBox.NoIcon, about_title, about_text)
        about_box.setTextFormat(Qt.RichText)
        about_box.setWindowModality(Qt.NonModal)
        help.addAction("&About Xi-CAM", lambda: about_box.show())

        menubar.addMenu(help)

        # Initialize layout with first plugin
        self._currentGUIPlugin = None
        self.build_layout()

        # self._currentGUIPlugin = pluginmanager.getPluginsOfCategory("GUIPlugin")[0]
        self.populate_layout()

        # Make F key bindings
        fkeys = [
            Qt.Key_F1,
            Qt.Key_F2,
            Qt.Key_F3,
            Qt.Key_F4,
            Qt.Key_F5,
            Qt.Key_F6,
            Qt.Key_F7,
            Qt.Key_F8,
            Qt.Key_F9,
            Qt.Key_F10,
            Qt.Key_F11,
            Qt.Key_F12,
        ]
        self.Fshortcuts = [QShortcut(QKeySequence(key), self) for key in fkeys]
        for i in range(12):
            self.Fshortcuts[i].activated.connect(partial(self.setStage, i))

        self.readSettings()
        # Wireup default widgets
        get_default_stage()["left"].sigOpen.connect(self.open)
        get_default_stage()["left"].sigOpen.connect(print)
        get_default_stage()["left"].sigPreview.connect(get_default_stage()["lefttop"].preview)

    def open(self, header):
        if self.currentGUIPlugin is None:
            msg.notifyMessage("Please select a gui plugin from the top before trying to open an image.")
            return
        if isinstance(header, Catalog):
            self.currentGUIPlugin.appendCatalog(header)
        elif isinstance(header, CatalogEntry):
            self.currentGUIPlugin.appendCatalog(header())
        elif isinstance(header, NonDBHeader):
            self.currentGUIPlugin.appendHeader(header)
        else:
            raise TypeError(f"Cannot open {header}.")

    def showSettings(self):
        ConfigDialog().show()

    def load_style(self):
        # Load styles explicitly
        AppearanceSettingsPlugin().apply()

    def setStage(self, stage):
        """
        Set the current Stage/Layout/Plugin mode to number i in its sequence. Triggered by menu (TODO) or F keybindings.

        Parameters
        ----------
        i   : int
        """
        plugin = self.currentGUIPlugin
        plugin.stage = stage
        self.populate_layout()

    def setGUIPlugin(self, guiplugin):
        self.currentGUIPlugin = guiplugin

    @property
    def currentGUIPlugin(self):
        return self._currentGUIPlugin

    @currentGUIPlugin.setter
    def currentGUIPlugin(self, guiplugin):
        if guiplugin != self._currentGUIPlugin:
            self._currentGUIPlugin = guiplugin
            self.populate_layout()

    def build_layout(self):
        # TODO: Allow save/restore https://stackoverflow.com/questions/14288635/any-easy-way-to-store-dock-widows-layout-and-sizes-in-settings-with-qt

        # Clear out docks
        for dockwidget in self.findChildren(QDockWidget):
            if dockwidget.parent() == self:
                self.removeDockWidget(dockwidget)

        # Make new docks
        self.topwidget = QDockWidget(parent=self)
        self.leftwidget = QDockWidget(parent=self)
        self.rightwidget = QDockWidget(parent=self)
        self.bottomwidget = QDockWidget(parent=self)
        self.lefttopwidget = QDockWidget(parent=self)
        self.righttopwidget = QDockWidget(parent=self)
        self.leftbottomwidget = QDockWidget(parent=self)
        self.rightbottomwidget = QDockWidget(parent=self)

        # Place the docks
        self.addDockWidget(Qt.LeftDockWidgetArea, self.lefttopwidget)
        self.addDockWidget(Qt.RightDockWidgetArea, self.righttopwidget)
        self.addDockWidget(Qt.TopDockWidgetArea, self.topwidget)
        self.addDockWidget(Qt.LeftDockWidgetArea, self.leftwidget)
        self.addDockWidget(Qt.RightDockWidgetArea, self.rightwidget)
        self.addDockWidget(Qt.BottomDockWidgetArea, self.bottomwidget)
        self.addDockWidget(Qt.LeftDockWidgetArea, self.leftbottomwidget)
        self.addDockWidget(Qt.RightDockWidgetArea, self.rightbottomwidget)

        # Adjust spanning
        self.setCorner(Qt.TopLeftCorner, Qt.LeftDockWidgetArea)
        self.setCorner(Qt.TopRightCorner, Qt.RightDockWidgetArea)
        self.setCorner(Qt.BottomLeftCorner, Qt.LeftDockWidgetArea)
        self.setCorner(Qt.BottomRightCorner, Qt.RightDockWidgetArea)

    def populate_layout(self):
        # Get current stage
        if self.currentGUIPlugin:
            stage = self.currentGUIPlugin.stage
        else:
            stage = get_default_stage()

        # Set center contents
        self.centralWidget().addWidget(stage.centerwidget)
        self.centralWidget().setCurrentWidget(stage.centerwidget)

        # Set visibility based on panel state and (TODO) insert default widgets when defaulted
        for position in ["top", "left", "right", "bottom", "lefttop", "righttop", "leftbottom", "rightbottom"]:
            self.populate_hidden(stage, position)
            self.populate_position(stage, position)

    def populate_hidden(self, stage, position):
        getattr(self, position + "widget").setHidden(
            (stage[position] == PanelState.Disabled)
            or (stage[position] == PanelState.Defaulted and get_default_stage()[position] == PanelState.Defaulted)
        )

    def populate_position(self, stage, position: str):
        if isinstance(stage[position], QWidget):
            getattr(self, position + "widget").setWidget(stage[position])
        elif stage[position] == PanelState.Defaulted:
            if not get_default_stage()[position] == PanelState.Defaulted:
                getattr(self, position + "widget").setWidget(get_default_stage()[position])
        elif isinstance(stage[position], type):
            raise TypeError(
                f"A type is not acceptable value for stages. You must instance this class: {stage[position]}, {position}"
            )

    def mousePressEvent(self, event):
        focused_widget = QApplication.focusWidget()
        if isinstance(focused_widget, QSpinBox):
            focused_widget.clearFocus()
        super(XicamMainWindow, self).mousePressEvent(event)

    def closeEvent(self, event):
        QSettings().setValue("geometry", self.saveGeometry())
        QMainWindow.closeEvent(self, event)

    def readSettings(self):
        settings = QSettings()
        if settings.value("geometry") is not None:
            self.restoreGeometry(settings.value("geometry"))


class Node(object):
    def __init__(self, object, name, children=None, parent=None):
        self.object = object
        self.name = name
        self.children = children or []
        self.parent = parent


class pluginModeWidget(QToolBar):
    """
    A series of styled QPushButtons with pipe characters between them. Used to switch between plugin modes.
    """

    sigSetStage = Signal(object)
    sigSetGUIPlugin = Signal(object)

    def __init__(self):
        super(pluginModeWidget, self).__init__()

        self.GUIPluginActionGroup = QActionGroup(self)

        # Setup font
        self.font = QFont("Zero Threes")
        self.font.setPointSize(16)

        # Align right
        self.setLayoutDirection(Qt.RightToLeft)

        # Subscribe to plugins
        pluginmanager.attach(self.pluginsChanged)

        self._nodes = []

        # Build children
        self.pluginsChanged()

    def pluginsChanged(self):
        self._build_nodes()
        self.showNode()

    def _build_nodes(self):
        self._nodes = []
        for plugin in pluginmanager.get_plugins_of_type("GUIPlugin") + pluginmanager.get_plugins_of_type("EZPlugin"):

            node = Node(plugin, plugin.name)

            for name, stage in plugin.stages.items():
                node.children.append(self._build_subnodes(name, stage, parent=node))

            self._nodes.append(node)

    def _build_subnodes(self, name, stage, parent):
        node = Node(stage, name, parent=parent)
        if isinstance(stage, dict):
            for subname, substage in stage.items():
                node.children.append(self._build_subnodes(subname, substage, parent=node))

        self._nodes.append(node)

        return node

    def showNode(self, node=None, direction=None):
        if node is None:  # toplevel
            plugin_nodes = [node for node in self._nodes if node.parent is None]
            # TODO: add plugin deactivation
            nodes = [node for node in plugin_nodes if getattr(node.object, "is_activated", True) or True]
            self._showNodes(nodes, direction)
            # self.fadeOut(callback=partial(self.mkButtons, names=names, callback=self.showStages), distance=0)

        elif isinstance(node.object, dict):
            nodes = node.children
            if len(nodes) > 1:
                self._showNodes(nodes, direction)

        elif isinstance(node.object, (PluginType,)):
            nodes = node.children
            if len(nodes) > 1:
                self._showNodes(nodes, direction)
            self.sigSetGUIPlugin.emit(node.object)
            self.setStage(node.object.stage)

        # elif isinstance(node.object, dict): # midlevel
        #     self._showNodes(node.object.keys(), node.object.values(), direction)

        else:  # bottomlevel
            parent_node = node.parent
            while parent_node.parent is not None:
                parent_node = parent_node.parent
            self.sigSetGUIPlugin.emit(parent_node.object)
            self.setStage(node.object)

            # self.fadeOut(callback=partial(self.mkButtons, names=names, callback=self.showStages), distance=0)

    def _showNodes(self, nodes, direction=None):
        if direction == "up":
            distance = 20
        elif direction == "down":
            distance = -20
        else:
            distance = 0

        self.fadeOut(callback=partial(self.mkButtons, nodes=nodes), distance=distance)

    def fadeOut(self, callback, distance=-20):
        duration = 200
        self._effects = []
        for action in self.actions():
            for widget in action.associatedWidgets():
                if widget is not self:
                    a = QPropertyAnimation(widget, b"pos", widget)
                    a.setStartValue(widget.pos())
                    a.setEndValue(widget.pos() + QPoint(0, distance))
                    self._effects.append(a)
                    a.setDuration(duration)
                    a.setEasingCurve(QEasingCurve.OutBack)
                    a.start(QPropertyAnimation.DeleteWhenStopped)

                    effect = QGraphicsOpacityEffect(self)
                    widget.setGraphicsEffect(effect)
                    self._effects.append(effect)
                    b = QPropertyAnimation(effect, b"opacity")
                    self._effects.append(b)
                    b.setDuration(duration)
                    b.setStartValue(1)
                    b.setEndValue(0)
                    b.setEasingCurve(QEasingCurve.OutBack)
                    b.start(QPropertyAnimation.DeleteWhenStopped)
                    b.finished.connect(partial(self.removeAction, action))
                    b.finished.connect(callback)
        if not self.actions():
            callback()

    def fadeIn(self):
        self._effects = []
        for action in self.actions():
            effect = QGraphicsOpacityEffect(self)
            self._effects.append(effect)
            for widget in action.associatedWidgets():
                if widget is not self:
                    widget.setGraphicsEffect(effect)
            a = QPropertyAnimation(effect, b"opacity")
            self._effects.append(a)
            a.setDuration(1000)
            a.setStartValue(0)
            a.setEndValue(1)
            a.setEasingCurve(QEasingCurve.OutBack)
            a.start(QPropertyAnimation.DeleteWhenStopped)

    def setStage(self, stage):
        self.sigSetStage.emit(stage)

    def mkButtons(self, nodes):
        # Remove+delete previous children
        layout = self.layout()
        for i in reversed(range(layout.count())):
            layout.itemAt(i).widget().setParent(None)

        if not nodes:
            return

        parent = nodes[0].parent

        if parent is not None:
            action = QAction("↑", self)
            action.setFont(self.font)
            action.triggered.connect(partial(self.showNode, parent.parent, direction="up"))
            action.setProperty("isMode", True)
            self.addAction(action)
            # Make separator pipe
            label = QAction("|", self)
            label.setFont(self.font)
            label.setDisabled(True)
            self.addAction(label)

        # Loop over each "GUIPlugin" plugin
        for node in reversed(list(nodes)):
            action = QAction(node.name, self)
            action.triggered.connect(partial(self.showNode, node, direction="down"))
            action.setFont(self.font)
            action.setProperty("isMode", True)
            action.setCheckable(True)
            action.setActionGroup(self.GUIPluginActionGroup)
            self.addAction(action)

            # Make separator pipe
            label = QAction("|", self)
            label.setFont(self.font)
            label.setDisabled(True)
            self.addAction(label)

        # Remove last separator
        if self.layout().count():
            self.layout().takeAt(self.layout().count() - 1).widget().deleteLater()  # Delete the last pipe symbol

        if parent is not None:
            # Make separator pipe
            label = QAction(">", self)
            label.setFont(self.font)
            label.setDisabled(True)
            self.addAction(label)

            action = QAction(parent.name, self)
            action.setFont(self.font)
            action.setProperty("isMode", True)
            action.setDisabled(True)
            action.setActionGroup(self.GUIPluginActionGroup)
            self.addAction(action)

        self.fadeIn()



from traitlets import TraitType
from traitlets.config.loader import PyFileConfigLoader, ConfigFileNotFound, Config

CONFIG_FILE_NAME = "bluesky_browser_config.py"
CONFIG_SEARCH_PATH = "."


def load_config():
    loader = PyFileConfigLoader(CONFIG_FILE_NAME, CONFIG_SEARCH_PATH)
    try:
        config = loader.load_config()
    except ConfigFileNotFound:
        config = Config()
    return config


class Callable(TraitType):
    """A trait which is callable.
    Notes
    -----
    Classes are callable, as are instances
    with a __call__() method."""

    info_text = "a callable"

    def validate(self, obj, value):
        if callable(value):
            return value
        else:
            self.error(obj, value)



from xicam.plugins.guiplugin import GUILayout
from .motd import MOTD
from .dataresourcebrowser import DataResourceBrowser
from .previewwidget import PreviewWidget

_defaultstage = None


def get_default_stage():
    global _defaultstage
    if not _defaultstage:
        _defaultstage = GUILayout(center=MOTD(), left=DataResourceBrowser(), lefttop=PreviewWidget())
    return _defaultstage



"""
Usage: pip install -e .
       python setup.py install
       python setup.py bdist_wheel
       python setup.py sdist bdist_egg
       twine upload dist/*
"""

# NOTE: Requires cx_freeze==6.0.b1 for linux builds

import sys
import versioneer
from codecs import open
from os import path

# These bits don't get collected automatically when packaging:
loosebits = ["numpy.core._methods", "numpy.lib.recfunctions"]

from setuptools import setup, find_packages, find_namespace_packages

# Set the long_description from the README
here = path.abspath(path.dirname(__file__))
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="xicam",
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="The CAMERA platform for synchrotron data management, visualization, and reduction.",
    long_description=long_description,
    # The project's main homepage.
    url="https://github.com/Xi-cam/Xi-cam",
    download_url="https://pypi.org/project/xicam/",
    # Author details
    author="Ronald J Pandolfi",
    author_email="ronpandolfi@lbl.gov",
    # Choose your license
    license="BSD",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Physics",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: BSD License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3.7",
    ],
    # What does your project relate to?
    keywords="synchrotron analysis x-ray scattering tomography ",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages() + find_namespace_packages(include=["xicam.*"]),
    package_dir={},
    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    py_modules=["xicam._version", "xicam.run_xicam", "xicam.run_headless"],
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=["PyQt5>=5.9.2",
                      "pathlib",
                      "qtpy",
                      "astropy",
                      "numpy>=1.16.0",
                      "appdirs",
                      "qdarkstyle",
                      "qtmodern",
                      "databroker>=1.0.0b8",
                      "distributed",
                      "dask",
                      "astropy",
                      "virtualenv",
                      "requests",
                      "appdirs",
                      "entrypoints",
                      "pyqtgraph",
                      "qtconsole", ],
    setup_requires=["cython"],  # cython is a missing required dep of astropy
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,tests]
    extras_require={
        "docs": ["sphinx", "recommonmark", "sphinx_bootstrap_theme", "sphinx-markdown-tables"],
        "tests": ["pytest", "coverage", "coveralls", "codecov", "pylint", "pytest-qt", "pytest-cov"],
        "pyqt5": ["pyqt5"],
        "PySide2": ["PySide2"],
    },
    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={},
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[#('lib/python2.7/site-packages/gui', glob.glob('gui/*')),
    #            ('lib/python2.7/site-packages/yaml/tomography',glob.glob('yaml/tomography/*'))],
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        "gui_scripts": ["xicam=xicam.run_xicam:main"],
        "xicam.plugins.DataHandlerPlugin": ["npy = xicam.core.formats.NPYPlugin:NPYPlugin"],
        "xicam.plugins.PluginType": [
            "CatalogPlugin = xicam.plugins.catalogplugin:CatalogPlugin",
            "ControllerPlugin = xicam.plugins.controllerplugin:ControllerPlugin",
            "DataHandlerPlugin = xicam.plugins.datahandlerplugin:DataHandlerPlugin",
            "DataResourcePlugin = xicam.plugins.dataresourceplugin:DataResourcePlugin",
            "EZPlugin = xicam.plugins.ezplugin:_EZPlugin",
            "Fittable1DModelPlugin = xicam.plugins.fittablemodelplugin:Fittable1DModelPlugin",
            "GUIPlugin = xicam.plugins.guiplugin:GUIPlugin",
            'OperationPlugin = xicam.plugins.operationplugin:OperationPlugin',
            "SettingsPlugin = xicam.plugins.settingsplugin:SettingsPlugin",
            "QWidgetPlugin = xicam.plugins.widgetplugin:QWidgetPlugin",
        ],
        "xicam.plugins.SettingsPlugin": [
            "logging = xicam.gui.settings.logging:LoggingSettingsPlugin",
            "appearance = xicam.gui.settings.appearance:AppearanceSettingsPlugin",
            "execution = xicam.gui.settings.execution:ExecutionSettingsPlugin",
            "connections = xicam.gui.connections:ConnectionSettingsPlugin",
            "cammart = xicam.gui.cammart:CamMartSettingsPlugin",
            "venvs = xicam.gui.cammart.venvs:VenvsSettingsPlugin",
        ],
        "xicam.plugins.CatalogPlugin": [
            "databroker_catalog_plugin = xicam.gui.bluesky.databroker_catalog_plugin:DatabrokerCatalogPlugin"
        ],
    },
    ext_modules=[],
    include_package_data=True,
)



from qtpy.QtWidgets import QLabel
from xicam.plugins import GUIPlugin, GUILayout


class TestPlugin(GUIPlugin):
    name = "test"

    def __init__(self):
        self.stages = {
            "Stage 1": GUILayout(QLabel("Stage 1")),
            "Stage 2": {"Stage 2.1": GUILayout(QLabel("Stage 2"))},
            "Stage 3": {"Stage 3.1": GUILayout(QLabel("Stage 3.1")), "Stage 3.2": GUILayout(QLabel("Stage 3.2"))},
        }

        super(TestPlugin, self).__init__()



# Modification of David Zuber's MenuView at https://github.com/storax/qmenuview and on PyPi
# Changes:
# - Uses qtpy instead of PySide

# Copyright (c) 2015, David Zuber
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# * Neither the name of Pyside QMenuView nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import functools

from qtpy import QtCore, QtGui, QtWidgets

__all__ = ["MenuView", "SetDataArgs"]


class MenuView(QtWidgets.QMenu):
    """A view that creates submenus based on a model.
    The model can be a list, table or treemodel.
    Each row equals to one submenu/action.
    In a treemodel, the leaves are plain actions, the rest also have menus on top.
    The view listens to the following signals:
      - :data:`PySide.QtCore.QAbstractItemModel.modelReset`
      - :data:`PySide.QtCore.QAbstractItemModel.rowsInserted`
      - :data:`PySide.QtCore.QAbstractItemModel.rowsAboutToBeRemoved`
      - :data:`PySide.QtCore.QAbstractItemModel.dataChanged`
    So the view is quite dynamic. If all child rows of an index are removed,
    the menu gets removed from the action. If rows are inserted to a parent index,
    which had no children, the action will get a menu.
    If an action emits a signal, the view will emit a signal too.
    The signal will contain the index and any arguments of the action's signal.
    You can get the action by using :meth:`MenuView.get_action`.
    See :data:`MenuView.action_triggered`, :data:`MenuView.action_hovered`,
    :data:`MenuView.action_toggled`.
    .. Note:: At the moment :data:`PySide.QtGui.QAction.changed` will not be handled.
              There currently is a bug in :meth:`MenuView._get_parents`, which
              causes an infinite loop.
    You can set which column to use for each attribute. See :data:`MenuView.text_column`,
    :data:`MenuView.icon_column`, :data:`MenuView.icontext_column`,
    :data:`MenuView.tooltip_column`, :data:`MenuView.checked_column`,
    :data:`MenuView.whatsthis_column`, :data:`MenuView.statustip_column`.
    For more control on how the data gets applied to the action, change
    :data:`MenuView.setdataargs`. It is a list of :class:`SetDataArgs` containers.
    One container defines the functionname to use for setting the attribute,
    the column to use, the :data:`PySide.QtCore.Qt.ItemDataRole`, and a data conversion function.
    If you want custom menu and action classes,
    override :meth:`MenuView.create_menu`, :meth:`MenuView.create_action`.
    """

    action_hovered = QtCore.Signal(QtCore.QModelIndex)
    """Signal for when an action gets hovered"""
    action_triggered = QtCore.Signal(QtCore.QModelIndex, bool)
    """Signal for when an action gets triggered"""
    action_toggled = QtCore.Signal(QtCore.QModelIndex, bool)
    """Signal for when an action gets toggled"""

    def __init__(self, title="", parent=None):
        """Initialize a new menu view with the given title
        :param title: title of the top menu
        :type title: :class:`str`
        :param parent: the parent widget
        :type parent: :class:`PySide.QtGui.QWidget`
        :raises: None
        """
        super(MenuView, self).__init__(title, parent)
        self.text_column = 0
        """The column for the action text. Default 0"""
        self.icon_column = 0
        """The column for the action icon. Default 0"""
        self.icontext_column = -1
        """The column for the action icon text. Default -1"""
        self.tooltip_column = 0
        """The column for the tooltip data. Default 0"""
        self.checked_column = 0
        """The column for the checked data. Has to be checkable. Default 0"""
        self.whatsthis_column = 0
        """The column for the whatsThis text. Default 0"""
        self.statustip_column = 0
        """The column for the statustip text. Default 0"""
        self._model = None

        Qt = QtCore.Qt
        args = [
            SetDataArgs("setText", "text_column", Qt.DisplayRole, str),
            SetDataArgs("setIcon", "icon_column", Qt.DecorationRole, self._process_icondata),
            SetDataArgs("setIconText", "icontext_column", Qt.DisplayRole, str),
            SetDataArgs("setToolTip", "tooltip_column", Qt.ToolTipRole, str),
            SetDataArgs("setChecked", "checked_column", Qt.CheckStateRole, self._checkconvertfunc),
            SetDataArgs("setWhatsThis", "whatsthis_column", Qt.WhatsThisRole, str),
            SetDataArgs("setStatusTip", "statustip_column", Qt.StatusTipRole, str),
        ]
        self.setdataargs = args
        """A list of :class:`SetDataArgs` containers. Defines how the
        data from the model is applied to the action"""

    @property
    def model(self,):
        """Get the model
        :returns: the current model
        :rtype: :class:`PySide.QtCore.QAbstractItemModel`
        :raises: None
        """
        return self._model

    @model.setter
    def model(self, model):
        """Set the model
        :param model: the model to set
        :type model: :class:`PySide.QtCore.QAbstractItemModel`
        :returns: None
        :rtype: None
        :raises: None
        """
        signalmap = {
            "modelReset": self.reset,
            "rowsInserted": self.insert_menus,
            "rowsAboutToBeRemoved": self.remove_menus,
            "dataChanged": self.update_menus,
        }
        if self._model:
            for signal, callback in signalmap.items():
                getattr(self._model, signal).disconnect(callback)
        self._model = model
        if model:
            for signal, callback in signalmap.items():
                getattr(model, signal).connect(callback)

        self.action_toggled.connect(functools.partial(self._model.setData, role=QtCore.Qt.CheckStateRole))
        self.reset()

    def reset(self,):
        """Delete and recreate all menus
        :returns: None
        :rtype: None
        :raises: None
        """
        self.clear()
        self.create_all_menus()

    def create_all_menus(self,):
        """Create all menus according to the model
        :returns: None
        :rtype: None
        :raises: None
        """
        m = self._model
        if not m:
            return
        indizes = self._flatten_hierarchy(m)
        for i in indizes:
            self.create_menu_for_index(i)

    @staticmethod
    def _flatten_hierarchy(model, parent=None):
        """Return a level-order list of indizes
        :param model: the model to traverse
        :type model: :class:`PySide.QtCore.QAbstractItemModel`
        :param parent: the parent index. Default is the root.
        :type parent: :class:`PySide.QtCore.QModelIndex`
        :returns: a level-order list of indizes
        :rtype: :class:`list` of :class:`PySide.QtCore.QModelIndex`
        :raises: None
        """
        indizes = []
        if parent is None:
            parent = QtCore.QModelIndex()
        parents = [parent]
        children = []
        # for all parents, get the children
        # during the next iteration, the children become the parents
        # and are queried for children as well. Once no new children are found
        # the loop ends
        while parents:
            for parent in parents:
                for i in range(model.rowCount(parent)):
                    index = model.index(i, 0, parent)
                    indizes.append(index)
                    children.append(index)
            parents = children
            children = []
        return indizes

    def create_menu_for_index(self, index):
        m = self._model
        parentaction = self.get_action(index.parent())
        # Action has no menu yet. In order to create a sub action,
        # we have to convert it.
        if parentaction.menu() is None:
            self._convert_action_to_menu(parentaction)
        parent = parentaction.menu()
        beforeindex = index.sibling(index.row(), 0)
        before = self.get_action(beforeindex)
        if m.hasChildren(index):
            action = self.create_menu(parent)
        else:
            action = self.create_action(parent)
        parent.insertAction(before, action)
        self.set_action_data(action, index)
        signalmap = {
            action.triggered: self._action_triggered,
            action.hovered: self._action_hovered,
            action.toggled: self._action_toggled,
        }
        for signal, callback in signalmap.items():
            signal.connect(functools.partial(callback, action))

    def _convert_action_to_menu(self, action):
        parent = action.parentWidget()
        menuaction = self.create_menu(parent)
        action.setMenu(menuaction.menu())

    def create_menu(self, parent):
        """Create a menu and return the menus action.
        The parent of the menu has to be set to ``parent``
        :param parent: The parent menu
        :type parent: :class:`PySide.QtGui.QMenu`
        :returns: The menu action
        :rtype: :class:`PySide.QtGui.QAction`
        :raises: None
        """
        menu = QtWidgets.QMenu(parent=parent)
        return menu.menuAction()

    def create_action(self, parent):
        """Create and return a new action
        The parent of the action has to be set to ``parent``
        :param parent: The parent menu
        :type parent: :class:`PySide.QtGui.QMenu`
        :returns: The created action
        :rtype: :class:`PySide.QtGui.QAction`
        :raises: None
        """
        return QtWidgets.QAction(parent)

    def insert_menus(self, parent, first, last):
        """Create menus for rows first til last under the given parent
        :param parent: The parent index
        :type parent: :class:`PySide.QtCore.QModelIndex`
        :param first: the first row
        :type first: :class:`int`
        :param last: the last row
        :type last: :class:`int`
        :returns: None
        :rtype: None
        :raises: None
        """
        for i in range(first, last + 1):
            index = self._model.index(i, 0, parent)
            flattened = [index]
            flattened.extend(self._flatten_hierarchy(self._model, index))
            for newi in flattened:
                self.create_menu_for_index(newi)

    def remove_menus(self, parent, first, last):
        """Remove the menus under the given parent
        :param parent: the parent of the menus
        :type parent: :class:`PySide.QtCore.QModelIndex`
        :param first: the first row
        :type first: :class:`int`
        :param last: the last row
        :type last: :class:`int`
        :returns: None
        :rtype: None
        :raises: None
        """
        parentaction = self.get_action(parent)
        parentmenu = parentaction.menu()
        for i in reversed(range(first, last + 1)):
            index = self._model.index(i, 0, parent)
            action = self.get_action(index)
            parentmenu.removeAction(action)
        # menu has no childs, only display the action
        if not parentmenu.actions() and parentmenu is not self:
            parentaction.setMenu(None)

    def update_menus(self, topLeft, bottomRight):
        """Update the menus from topleft index to bottomright index
        :param topLeft: The top left index to update
        :type topLeft: :class:`PySide.QtCore.QModelIndex`
        :param bottomRight: the bottom right index to update
        :type bottomRight: :class:`PySide.QtCore.QModelIndex`
        :returns: None
        :rtype: None
        :raises: None
        """
        columns = [
            self.text_column,
            self.icon_column,
            self.icontext_column,
            self.tooltip_column,
            self.checked_column,
            self.whatsthis_column,
            self.statustip_column,
        ]
        needupdate = any([(c >= topLeft.column() and c <= bottomRight.column()) for c in columns])
        if needupdate:
            for row in range(topLeft.row(), bottomRight.row() + 1):
                index = topLeft.sibling(row, 0)
                action = self.get_action(index)
                self.set_action_data(action, index)

    def get_index(self, action, column=0):
        """Return the index for the given action
        :param action: the action to query
        :type action: :class:`PySide.QtGui.QAction`
        :param column: The column of the index
        :type column: :class:`int`
        :returns: the index of the action
        :rtype: :class:`PySide.QtCore.QModelIndex`
        :rasies: None
        """
        if action == self.menuAction():
            return QtCore.QModelIndex()
        # find all parents to get their index
        parents = self._get_parents(action)
        index = QtCore.QModelIndex()
        # Move through indexes down the chain
        for a in reversed(parents):
            parent = a.parent()
            # if parent of action is its own menu, get parent of that menu.
            # We want to know which row the action is in. For that we need
            # The real parent menu.
            if parent is a.menu():
                parent = parent.parent()
            row = parent.actions().index(a)
            index = self._model.index(row, 0, index)
        parent = action.parent()
        if parent is None:
            return index
        if parent is action.menu():
            parent = parent.parent()
        row = parent.actions().index(action)
        index = self._model.index(row, column, index)
        return index

    def _get_parents(self, action):
        parents = []
        a = action
        while True:
            parent = a.parent()
            # if the parent is the actions menu, we have to get
            # the menues parent. Else we get stuck on the same level.
            if parent and parent is a.menu():
                parent = parent.parent()
            if not isinstance(parent, QtWidgets.QMenu):
                # Is not part of the tree. Parent is not a menu but
                # might be None or another Widget
                return []
            # break if parent is root because we got all parents we need
            if parent is self:
                break
            # a new parent was found and we are still not at root
            # search further until we get to root
            parent = parent.menuAction()
            a = parent
            parents.append(parent)
        return parents

    def get_action(self, index):
        """Return the action for the given index
        :param index: the index to query
        :type index: :class:`PySide.QtCore.QModelIndex`
        :returns: the action for the given index
        :rtype: :class:`PySide.QtGui.QAction`
        :raises: None
        """
        if not index.isValid():
            return self.menuAction()
        parents = self._get_parent_indizes(index)
        menu = self
        for i in reversed(parents):
            action = menu.actions()[i.row()]
            menu = action.menu()
        try:
            return menu.actions()[index.row()]
        except IndexError:
            return None

    def _get_parent_indizes(self, index):
        if not index.isValid() or index.model() != self._model:
            return []
        parents = []
        i = index
        while True:
            p = i.parent()
            if not p.isValid():
                break
            parents.append(p)
            i = p
        return parents

    def set_action_data(self, action, index):
        """Set the data of the action for the given index
        .. Note:: The column of the index does not matter. The columns for the data
                  are specified in :data:`MenuView.setdataargs`.
        The arguments to used are defined in :data:`MenuView.setdataargs`.
        :param action: The action to update
        :type action: :class:`PySide.QtGui.QAction`
        :param index: The index with the data
        :type index: :class:`PySide.QtCore.QModelIndex`
        :returns: None
        :rtype: None
        :raises: None
        """
        self._set_action_enabled(action, index)
        self._set_action_checkable(action, index)
        for args in self.setdataargs:
            self._set_action_attribute(action, index, args)

    def _set_action_enabled(self, action, index):
        """Enable the action , depending on the item flags
        :param action: The action to update
        :type action: :class:`PySide.QtGui.QAction`
        :param index: the model index with the item flags
        :type index: :class:`PySide.QtCore.QModelIndex`
        :returns: None
        :rtype: None
        :raises: None
        """
        action.setEnabled(index.flags() & QtCore.Qt.ItemIsEnabled)

    def _set_action_checkable(self, action, index):
        """Set the action checkable, depending on the item flags
        .. Note:: The column of the index does not matter. The column for the data
                  is specified in :data:`MenuView.checked_column`.
        :param action: The action to update
        :type action: :class:`PySide.QtGui.QAction`
        :param index: the model index with the item flags
        :type index: :class:`PySide.QtCore.QModelIndex`
        :returns: None
        :rtype: None
        :raises: None
        """
        checkedindex = index.sibling(index.row(), self.checked_column)
        checkedflags = checkedindex.flags()
        action.setCheckable(checkedflags & QtCore.Qt.ItemIsUserCheckable)

    def _set_action_attribute(self, action, index, setdataarg):
        """Query the data of index and use it to set an attribute on action.
        .. Note:: The column of the index does not matter. The column for the data
                  is specified in ``setdataarg.column``.
        The data will be converted with ``setdataarg.convertfunc``.
        :param action: the action to update
        :type action: :class:`PySide.QtGui.QAction`
        :param index: the index with the data
        :type index: :class:`PySide.QtCore.QModelIndex`
        :param setdataarg: The container with arguments that define the way the data is applied
        :type setdataarg: :class:`SetDataArgs`
        :returns: None
        :rtype: None
        :raises: None
        """
        if isinstance(setdataarg.column, int):
            column = setdataarg.column
        else:
            column = getattr(self, setdataarg.column)
        data = self.get_data(index, setdataarg.role, column)
        if data is None:
            return
        setattrmethod = getattr(action, setdataarg.setfunc)
        if setdataarg.convertfunc:
            data = setdataarg.convertfunc(data)
        setattrmethod(data)

    @staticmethod
    def get_data(index, role, column=None):
        """Get data of the given index
        If the column is is not None and different from
        the index column, will get the data from a sibling
        index with the same row.
        :param index: The index to query for data
        :type index: :class:`PySide.QtCore.QModelIndex`
        :param role: the data role
        :type role: :data:`PySide.QtCore.Qt.ItemDataRole`
        :param column: the column of the row to query for data.
        :type column: :class:`int` | None
        :returns: The data retrieved
        :raises: None
        """
        if column and index.column() != column:
            index = index.sibling(index.row(), column)
        if index.isValid():
            return index.data(role)

    def _action_hovered(self, action):
        """Emit the hovered signal
        :param action: The action which emitted a hovered signal
        :type action: :class:`PySide.QtGui.QAction`
        :returns: None
        :rtype: None
        :raises: None
        """
        self._emit_signal_for_action(self.action_hovered, action)

    def _action_triggered(self, action, checked=False):
        """Emit the triggered signal
        :param action: The action which emitted a triggered signal
        :type action: :class:`PySide.QtGui.QAction`
        :param checked: True if the action was in a checked state
        :type checked: :class:`bool`
        :returns: None
        :rtype: None
        :raises: None
        """
        self._emit_signal_for_action(self.action_triggered, action, checked)

    def _action_toggled(self, action, checked=False):
        """Emit the toggled signal
        :param action: The action which emitted a toggled signal
        :type action: :class:`PySide.QtGui.QAction`
        :param checked: True if the action was in a checked state
        :type checked: :class:`bool`
        :returns: None
        :rtype: None
        :raises: None
        """
        self._emit_signal_for_action(self.action_toggled, action, checked)

    def _emit_signal_for_action(self, signal, action, *args):
        """Emit the given signal for the index of the given action
        :param signal: The signal to emit
        :type signal: :class:`PySide.QtCore.Signal`
        :param action: The action for which to emit the signal
        :type action: :class:`PySide.QtGui.QAction`
        :returns: None
        :rtype: None
        :raises: None
        """
        index = self.get_index(action)
        if index and index.isValid():
            signal.emit(index, *args)

    @staticmethod
    def _process_icondata(icondata):
        """Return an icon for the data of the :data:`PySide.QtCore.Qt.DecorationRole`
        :param icondata: The data from the :data:`PySide.QtCore.Qt.DecorationRole`
        :type icondata: :class:`PySide.QtGui.QIcon` | :class:`PySide.QtGui.QPixmap`
        :returns: A Icon based on the data.
        :rtype: :class:`PySide.QtGui.QIcon`
        :raises: None
        """
        if isinstance(icondata, QtGui.QIcon):
            return icondata
        if isinstance(icondata, QtGui.QPixmap):
            return QtGui.QIcon(icondata)

    @staticmethod
    def _checkconvertfunc(data):
        checkedstate = int(data) if data is not None else 0
        return checkedstate != QtCore.Qt.Unchecked


class SetDataArgs(object):
    """A container of arguments for setting attributes on an action.
    The data is queried from the model with ``role``. Then converted with ``convertfunc``.
    Then ``setfunc`` is used for setting the attribute on the action.
    ``convertfunc`` can be ``None``.
    If column is a string, the attribute of the view with that name will be used as column.
    """

    def __init__(self, setfunc, column, role, convertfunc):
        """Initialize a new container
        :raises: None
        """
        super(SetDataArgs, self).__init__()
        self.setfunc = setfunc
        self.column = column
        self.role = role
        self.convertfunc = convertfunc



import requests
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
from xicam.gui.static import path
from xicam.gui.widgets.searchlineedit import SearchLineEdit
from copy import deepcopy

from xicam.plugins import SettingsPlugin, manager


class ConnectionSettingsPlugin(SettingsPlugin):
    """
    A built-in settings plugin to configure connections to other hosts
    """

    def __init__(self):
        # Setup UI
        self.widget = QWidget()
        self.widget.setLayout(QHBoxLayout())
        self.listview = QListView()
        self.connectionsmodel = QStandardItemModel()
        self.listview.setModel(self.connectionsmodel)

        self.plugintoolbar = QToolBar()
        self.plugintoolbar.setOrientation(Qt.Vertical)
        self.plugintoolbar.addAction(QIcon(str(path("icons/plus.png"))), "Add plugin", self.add_credential)
        self.plugintoolbar.addAction(QIcon(str(path("icons/minus.png"))), "Remove plugin", self.remove_credential)
        self.widget.layout().addWidget(self.listview)
        self.widget.layout().addWidget(self.plugintoolbar)
        super(ConnectionSettingsPlugin, self).__init__(QIcon(str(path("icons/server.png"))), "Connections", self.widget)

    def add_credential(self):
        """
        Open the CamMart install dialog
        """
        self._dialog = CredentialDialog()
        self._dialog.sigAddCredential.connect(self._add_credential)
        self._dialog.exec_()

    def remove_credential(self):
        """
        Removes a credential
        """
        if self.listview.selectedIndexes():
            self.connectionsmodel.removeRow(self.listview.selectedIndexes()[0].row())

    def _add_credential(self, name: str, credential: dict):
        item = QStandardItem(name)
        item.credential = credential
        item.name = name
        self.connectionsmodel.appendRow(item)
        self.connectionsmodel.dataChanged.emit(item.index(), item.index())

    def toState(self):
        credentials = deepcopy(self.credentials)
        for name, credential in credentials.items():
            if credential.get("savepassword", False):
                credential["password"] = None
        return credentials

    def fromState(self, state):
        self.connectionsmodel.clear()
        for name, credential in state.items():
            item = QStandardItem(name)
            item.credential = credential
            item.name = name
            self.connectionsmodel.appendRow(item)
        self.listview.reset()

    @property
    def credentials(self):
        return {
            self.connectionsmodel.item(i).name: self.connectionsmodel.item(i).credential
            for i in range(self.connectionsmodel.rowCount())
        }


class CredentialDialog(QDialog):
    sigAddCredential = Signal(str, dict)
    sigConnect = Signal(dict)

    def __init__(self, addmode=True):
        super(CredentialDialog, self).__init__()

        # Set size and position
        # self.setGeometry(0, 0, 800, 500)
        frameGm = self.frameGeometry()
        screen = QApplication.desktop().screenNumber(QApplication.desktop().cursor().pos())
        centerPoint = QApplication.desktop().screenGeometry(screen).center()
        frameGm.moveCenter(centerPoint)
        self.move(frameGm.topLeft())

        # Setup fields
        self.host = QLineEdit()
        self.profiles = QComboBox()
        self.profiles.addItem("New...")
        self.profilename = QLineEdit()
        self.username = QLineEdit()
        self.password = QLineEdit()
        self.password.setEchoMode(QLineEdit.Password)
        self.savepassword = QCheckBox("Save Password")
        self.rememberprofile = QCheckBox("Remember Profile")

        # Setup dialog buttons
        self.addButton = QPushButton("&Add")
        self.connectButton = QPushButton("C&onnect")
        self.cancelButton = QPushButton("&Cancel")
        self.addButton.clicked.connect(self.add)
        self.connectButton.clicked.connect(self.connect)
        self.cancelButton.clicked.connect(self.close)
        self.profiles.currentTextChanged.connect(self.loadProfile)
        self.buttonboxWidget = QDialogButtonBox()
        if addmode:
            self.buttonboxWidget.addButton(self.addButton, QDialogButtonBox.AcceptRole)
        else:
            self.buttonboxWidget.addButton(self.connectButton, QDialogButtonBox.AcceptRole)

        self.buttonboxWidget.addButton(self.cancelButton, QDialogButtonBox.RejectRole)

        # Compose main layout
        mainLayout = QFormLayout()
        if not addmode:
            mainLayout.addRow("Profile", self.profiles)
        mainLayout.addRow("Profile", self.profilename)
        mainLayout.addRow("Host", self.host)
        mainLayout.addRow("Username", self.username)
        mainLayout.addRow("Password", self.password)
        mainLayout.addRow(self.savepassword)
        if not addmode:
            mainLayout.addRow(self.rememberprofile)
        mainLayout.addRow(self.buttonboxWidget)

        # Populate profiles
        for name, credential in manager.get_plugin_by_name("Connections", "SettingsPlugin").credentials.items():
            self.profiles.addItem(name)

        self.setLayout(mainLayout)
        self.setWindowTitle("Add Connection...")

        # Set modality
        self.setModal(True)

    def loadProfile(self):
        profilename = self.profiles.currentText()
        if profilename == "New...":
            self.username.setEnabled(True)
            self.password.setEnabled(True)
            self.host.setEnabled(True)
            self.savepassword.setEnabled(True)
            self.rememberprofile.setVisible(True)
        else:
            credential = manager.get_plugin_by_name("Connections", "SettingsPlugin").credentials[profilename]
            self.username.setText(credential["username"])
            self.host.setText(credential["host"])
            self.password.setText(credential["password"])
            self.savepassword.setChecked(credential["savepassword"])
            self.profilename.setText(profilename)
            self.username.setEnabled(False)
            self.password.setEnabled(False)
            self.host.setEnabled(False)
            self.savepassword.setEnabled(False)
            self.rememberprofile.setVisible(False)

    def add(self):
        self.sigAddCredential.emit(
            self.profilename.text(),
            {
                "host": self.host.text(),
                "username": self.username.text(),
                "password": self.password.text(),
                "savepassword": False,
            },
        )
        self.accept()

    def connect(self):
        if self.rememberprofile.isChecked():
            self.add()
        self.sigConnect.emit(
            {
                "host": self.host.text(),
                "username": self.username.text(),
                "password": self.password.text(),
                "savepassword": False,
            }
        )
        self.accept()  # Segfault?


class ConnectDelegate(QItemDelegate):
    def __init__(self, parent):
        super(ConnectDelegate, self).__init__(parent)
        self._parent = parent

    def paint(self, painter, option, index):
        if not self._parent.indexWidget(index):
            button = QToolButton(self.parent())
            button.setAutoRaise(True)
            button.setText("Delete Operation")
            button.setIcon(QIcon(path("icons/trash.png")))
            sp = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
            sp.setWidthForHeight(True)
            button.setSizePolicy(sp)
            button.clicked.connect(index.data())

            self._parent.setIndexWidget(index, button)



"""
Provides a central widget that contains much of what the blueskybrowser.
Might go away eventually, but needed to prevent the CentralWdiget from
trying to setup a menu.
"""

import time

from .search import SearchWidget, SearchState
from xicam.gui.bluesky.summary import SummaryWidget

from qtpy.QtCore import QDateTime, Qt
from qtpy.QtWidgets import QWidget, QHBoxLayout, QSplitter


class CentralWidget(QWidget):
    """
    Encapsulates all widgets and models. Connect signals on __init__.
    """

    def __init__(self, *args, catalog, menuBar, zmq_address=None, **kwargs):
        super().__init__(*args, **kwargs)

        # Define models.
        search_state = SearchState(catalog=catalog)
        self.search_model = search_state.search_results_model
        # Define widgets.
        self.search_widget = SearchWidget()
        self.summary_widget = SummaryWidget()

        left_pane = QSplitter(Qt.Vertical)
        left_pane.addWidget(self.search_widget)
        left_pane.addWidget(self.summary_widget)

        layout = QHBoxLayout()
        splitter = QSplitter(Qt.Horizontal)
        layout.addWidget(splitter)
        splitter.addWidget(left_pane)

        self.setLayout(layout)

        def show_double_clicked_entry(index):
            search_state.search_results_model.emit_open_entries(None, [index])

        def preview_clicked_entry(index):
            search_state.search_results_model.emit_preview_entry(None, index)

        # Set models, connect signals, and set initial values.
        now = time.time()
        ONE_WEEK = 60 * 60 * 24 * 7
        self.search_widget.search_results_widget.setModel(search_state.search_results_model)
        self.search_widget.search_input_widget.search_bar.textChanged.connect(
            search_state.search_results_model.on_search_text_changed
        )
        self.search_widget.catalog_selection_widget.catalog_list.setModel(search_state.catalog_selection_model)
        self.search_widget.search_input_widget.until_widget.dateTimeChanged.connect(
            search_state.search_results_model.on_until_time_changed
        )
        self.search_widget.search_input_widget.until_widget.setDateTime(QDateTime.fromSecsSinceEpoch(now + ONE_WEEK))
        self.search_widget.search_input_widget.since_widget.dateTimeChanged.connect(
            search_state.search_results_model.on_since_time_changed
        )
        self.search_widget.search_input_widget.since_widget.setDateTime(QDateTime.fromSecsSinceEpoch(now - ONE_WEEK))
        self.search_widget.catalog_selection_widget.catalog_list.currentIndexChanged.connect(
            search_state.set_selected_catalog
        )
        self.search_widget.search_results_widget.selectionModel().selectionChanged.connect(
            search_state.search_results_model.emit_selected_result
        )
        self.search_widget.search_results_widget.doubleClicked.connect(show_double_clicked_entry)
        self.search_widget.search_results_widget.clicked.connect(preview_clicked_entry)
        search_state.search_results_model.selected_result.connect(self.summary_widget.set_entries)
        search_state.search_results_model.valid_custom_query.connect(self.search_widget.search_input_widget.mark_custom_query)
        search_state.sig_update_header.connect(self.search_widget.search_results_widget.hide_hidden_columns)
        search_state.enabled = True
        search_state.search()



from astropy.modeling import Fittable1DModel
from .plugin import PluginType


class Fittable1DModelPlugin(Fittable1DModel, PluginType):
    is_singleton = False
    needs_qt = False

    """
    Plugins of this base class mimic the astropy FittableModel class structure. An activated fittable model would be
    usable for fitting 1-d spectra or 2-d images. Example: A 1-D Lorentzian model, usable for fitting SAXS spectra.

    See the Astropy API for a detailed explanation of the usage.

    See xicam.plugins.tests for examples.

    """

    def __init__(self, *args, **kwargs):
        super(Fittable1DModelPlugin, self).__init__(*args, **kwargs)

    @staticmethod
    def evaluate(x, *args, **kwargs):
        raise NotImplementedError

    @staticmethod
    def fit_deriv(x, *args, **kwargs):
        raise NotImplementedError

    @property
    def inverse(self):
        raise NotImplementedError

        # TODO: add fitting convenience method



from databroker.core import BlueskyRun
from pyqtgraph import ImageItem, TextItem, GraphicsLayoutWidget
import numpy as np
from qtpy.QtCore import QSize
from qtpy.QtGui import QFont, QTransform
from qtpy.QtWidgets import QSizePolicy
from xicam.core import msg, threads
from xicam.core.data import bluesky_utils

from xicam.core.data import NonDBHeader


class PreviewWidget(GraphicsLayoutWidget):
    def __init__(self):
        super(PreviewWidget, self).__init__()
        self.setMinimumHeight(250)
        self.setMinimumWidth(250)
        self.view = self.addViewBox(lockAspect=True, enableMenu=False)
        self.imageitem = ImageItem()
        self.textitem = TextItem(anchor=(0.5, 0))
        self.textitem.setFont(QFont("Zero Threes"))
        self.imgdata = None

        self.imageitem.setOpts(axisOrder="row-major")

        self.view.addItem(self.imageitem)
        self.view.addItem(self.textitem)
        self.textitem.hide()
        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        # def textItemBounds(axis, frac=1.0, orthoRange=None):
        #     b = self.textitem.boundingRect()
        #     sx, sy = self.view.viewPixelSize()
        #     x, y = sx*b.width(), sy*b.height()
        #     if axis == 0: return (-x/2, x/2)
        #     if axis == 1: return (0, y)
        #
        # self.textitem.dataBounds = textItemBounds

    def sizeHint(self):
        return QSize(250, 250)

    @threads.method(threadkey="preview", showBusy=False)
    def preview(self, data):
        if isinstance(data, NonDBHeader):
            self.preview_header(data)
        else:
            self.preview_catalog(data)

    @staticmethod
    def guess_stream_field(catalog: BlueskyRun):
        # TODO: use some metadata (techniques?) for guidance about how to get a preview

        for stream in ["primary", *bluesky_utils.streams_from_run(catalog)]:
            descriptor = bluesky_utils.descriptors_from_stream(catalog, stream)[0]
            fields = bluesky_utils.fields_from_descriptor(descriptor)
            for field in fields:
                field_ndims = bluesky_utils.ndims_from_descriptor(descriptor, field)
                if field_ndims > 1:
                    return stream, field

    def preview_catalog(self, catalog: BlueskyRun):
        threads.invoke_in_main_thread(self.setText, "LOADING...")
        try:
            stream, field = self.guess_stream_field(catalog)
            data = getattr(catalog, stream).to_dask()[field].squeeze()
            for i in range(len(data.shape) - 2):
                data = data[0]
            threads.invoke_in_main_thread(self.setImage, np.asarray(data.compute()))
        except Exception as ex:
            msg.logError(ex)
            threads.invoke_in_main_thread(self.imageitem.clear)
            threads.invoke_in_main_thread(self.setText, "UNKNOWN DATA FORMAT")

    def preview_header(self, header: NonDBHeader):
        try:
            data = header.meta_array()[0]
            threads.invoke_in_main_thread(self.setImage, data)
        except IndexError:
            threads.invoke_in_main_thread(self.imageitem.clear)
            threads.invoke_in_main_thread(self.setText, "UNKNOWN DATA FORMAT")

    def setImage(self, imgdata):
        self.imageitem.clear()
        self.textitem.hide()
        self.imgdata = imgdata
        self.imageitem.setImage(np.log(self.imgdata * (self.imgdata > 0) + (self.imgdata < 1)), autoLevels=True)
        self.imageitem.setTransform(QTransform(1, 0, 0, -1, 0, self.imgdata.shape[-2]))
        self.view.autoRange()

    def setText(self, text):
        self.textitem.setText(text)
        self.imageitem.clear()
        self.textitem.setVisible(True)
        self.view.autoRange()
